/*
 * idea.c
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 */

#include "xkernel.h"
#include "idea.h"
#include "idea_i.h"
#include "eth.h"		/* So ARP can go on top of CRYPT */
#include "ip.h"                /* For IP pseudoheader fixup. */
#include "km.h"
#include "saidmgr.h"
#include "msg.h"

#define maxInterleave  1024

#define TCRYPT(X) idea ## X
#define TCRYPTP ideap
#define TCRYPTS "idea"
#define MINKEYSZ 1
#define MAXKEYSZ 128

extern int traceideap; int traceideap;


/*  External routines:
    xk* are replacements for Sun library functionality, preserving some odd design choices.
    Address alignment note:  The data buffer addresses must be word aligned (addresses
     a multiple of 4 bytes, unless the machine architecture will handle load/store of
     words on non-aligned addresses.  Presently, the crypt routines always allocate
     their own working arrays, so the alignment is guaranteed by xMalloc, etc.
    Byte order note:  All data is grouped in the order ABCD EFGH for DES processing.
    On Little-Endian machines, a byte-swap+unswap is wrapped around each en/decryption.

 xk_idea_crypt  Does IDEA en/decryption on the whole message 
    unsigned char keystr[]  the key got from kms, must be char *
    unsigned char *data     the whole message to be en/decrypted
    long len                 length of the message buf
    long *mode               mode of IDEA
    int interleave          something like IV
*/

#ifdef __STDC__
int  xk_idea_crypt(unsigned char [], unsigned char *, long, long *, int);
void Error(int num, char *str);
#else
int  xk_idea_crypt();
void Error();
#endif /* __STDC__ */


#ifdef __STDC__

			void TCRYPT(_init)( XObj self );
static long		TCRYPT(copy)( void *, char *, long, void * );
static XObj		TCRYPT(CreateSessn)( XObj, XObj, XObj, XObj );
static void		TCRYPT(SessnInit)( XObj );
static int		TCRYPT(ControlProtl)( XObj, int, char *, int );
static xkern_return_t	TCRYPT(Demux)( XObj, XObj, Msg * );
static XObj		TCRYPT(Open)( XObj, XObj, XObj, Part * );
static xkern_return_t	TCRYPT(OpenEnable)( XObj, XObj, XObj, Part * );
static xkern_return_t	TCRYPT(OpenDisable)( XObj, XObj, XObj, Part * );
static xkern_return_t	TCRYPT(OpenDone)( XObj, XObj, XObj, XObj);
static xkern_return_t	TCRYPT(Close)( XObj );
static xmsg_handle_t	TCRYPT(Push)( XObj, Msg * );
static xkern_return_t	TCRYPT(Pop)( XObj, XObj, Msg *, void * );
static void 		TCRYPT(Store)( void *, char *, long, void * );
static int		TCRYPT(ControlSessn)( XObj, int, char *, int );
static int		xCopyParts(Part **, char **, Part *, XObj);
static void		xSimpleCopyParts(Part **, Part *);
static void             hexbprint(u_char *, int);
static void             print_crypt_session_state(char*,SState*,long);
static XObj             cryptipseckmOpen(Part *, XObj, int);

#else

			void TCRYPT(_init)();
static XObj		TCRYPT(CreateSessn)();
static void		TCRYPT(SessnInit)();
static int		TCRYPT(ControlProtl)();
static xkern_return_t	TCRYPT(Demux)();
static XObj		TCRYPT(Open)();
static xkern_return_t	TCRYPT(OpenEnable)();
static xkern_return_t	TCRYPT(OpenDisable)();
static xkern_return_t	TCRYPT(OpenDone)();
static xkern_return_t	TCRYPT(Close)();
static xmsg_handle_t	TCRYPT(Push)();
static xkern_return_t	TCRYPT(Pop)();
static int		TCRYPT(ControlSessn)();
static int		xCopyParts();
static void		xSimpleCopyParts();
static void             hexbprint();
static void             print_crypt_session_state();
static XObj             cryptipseckmOpen( );

#endif /* __STDC__ */

static void                /* print a byte string in hex */
hexbprint(buf,len) u_char *buf;int len;
{ for (;len>0;len--,buf++) printf(" %02x",*buf); }

static void                /* print a CRYPT session state */
print_crypt_session_state(label,ss,pflags) char *label; SState *ss; long pflags;
{  printf("%s  enable %x  kms %x  kms2 %x  part0 %x  part1 %x  method %x  flags %x  pflags %x\n",
     label,ss->e,ss->kms,ss->kms2,ss->p[0],ss->p[1],ss->method,ss->flags,pflags);
   printf("\n");  }

static xkern_return_t romfindpartnum(XObj, char **, int, int, void *);
static xkern_return_t romuseipsec(XObj, char **, int, int, void *);
static xkern_return_t romPhoturis(XObj, char **, int, int, void *);
static xkern_return_t romrfc1829(XObj, char **, int, int, void *);

static xkern_return_t romkeyformat(XObj, char **, int, int, void *);
static xkern_return_t romnumround(XObj, char **, int, int, void *);

static xkern_return_t TCRYPT(parameter_EECB)( XObj, char **, int, int, void * );
static xkern_return_t TCRYPT(parameter_DECB)( XObj, char **, int, int, void * );
static xkern_return_t TCRYPT(parameter_ECBC)( XObj, char **, int, int, void * );
static xkern_return_t TCRYPT(parameter_DCBC)( XObj, char **, int, int, void * );
static xkern_return_t TCRYPT(parameter_ECFB)( XObj, char **, int, int, void * );
static xkern_return_t TCRYPT(parameter_DCFB)( XObj, char **, int, int, void * );
static xkern_return_t TCRYPT(parameter_EOFB)( XObj, char **, int, int, void * );
static xkern_return_t TCRYPT(parameter_DOFB)( XObj, char **, int, int, void * );


static XObjRomOpt rmopts[] = {{ "numparts", 1, romfindpartnum }
			     ,{ "useipsec", 0, romuseipsec    }
			     ,{ "Photuris", 0, romPhoturis }
			     ,{ "rfc1829", 0, romrfc1829 }

			     ,{ "keyformat", 1, romkeyformat   }
			     ,{ "numround",  1, romnumround    }
 
			     ,{ "EECB", 0, TCRYPT(parameter_EECB)}
			     ,{ "DECB", 0, TCRYPT(parameter_DECB)}
			     ,{ "ECBC", 0, TCRYPT(parameter_ECBC)}
			     ,{ "DCBC", 0, TCRYPT(parameter_DCBC)}
			     ,{ "ECFB", 0, TCRYPT(parameter_ECFB)}
			     ,{ "DCFB", 0, TCRYPT(parameter_DCFB)}
			     ,{ "EOFB", 0, TCRYPT(parameter_EOFB)}
			     ,{ "DOFB", 0, TCRYPT(parameter_DOFB)}
};

static xkern_return_t
romrfc1829(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
    printf("\nromrfc1829\n"); return XK_FAILURE;

    ((PState *)(self->state))->flags |= RFC_1829;
    return XK_SUCCESS;
}

static xkern_return_t
romPhoturis(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
    printf("\nromPhoturis\n"); return XK_FAILURE;

    ((PState *)(self->state))->flags |= SEND_IV | IV_FROM_MSG |
			 IV_ON_UPMSG | NO_PADDING_CHECK | RFC_1829;
    return XK_SUCCESS;
}

static xkern_return_t
romuseipsec(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
    ((PState *)(self->state))->flags |= USE_IPSEC;
    return XK_SUCCESS;
}

static xkern_return_t
romfindpartnum(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  if (atoi(ptr[2]) > 1)
    ((PState *)(self->state))->flags |= ALL_PARTS;
  else
    ((PState *)(self->state))->flags &= ~ALL_PARTS;
  return XK_SUCCESS;
}

static xkern_return_t	
romkeyformat( self, ptr, fields, line, arg )
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  if (strcmp( ptr[2], "charstring" ))
    ((PState *)(self->state))->flags |= HEXSTRING;
  else
    ((PState *)(self->state))->flags &= ~HEXSTRING;
  return XK_SUCCESS;
  /* default value: CHAR STRING */
}

static xkern_return_t
romnumround(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  if (atoi(ptr[2]) >= 1)
    ((PState *)(self->state))->interleave = atoi( ptr[2] );
  else
    Error( 13, "interleave factor is zero" );
  if (((PState *)(self->state))->interleave >= maxInterleave )
    Error( 12, "interleave factor is too large" );
  return XK_SUCCESS;
  /* default value: 1 */
}


static xkern_return_t	
TCRYPT(parameter_EECB)( self, ptr, fields, line, arg )
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS " EECB mode");
  ((PState *)(self->state))->flags |= ECB;
  ((PState *)(self->state))->flags &= ~DECRYPT;
  return XK_SUCCESS;
  /* default value: Encrypt */
}

static xkern_return_t
TCRYPT(parameter_DECB)( self, ptr, fields, line, arg )
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS " DECB mode");
  ((PState *)(self->state))->flags |= ECB;
  ((PState *)(self->state))->flags |= DECRYPT;
  return XK_SUCCESS;
}

static xkern_return_t	
TCRYPT(parameter_ECBC)( self, ptr, fields, line, arg )
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS " ECBC mode");
  ((PState *)(self->state))->flags |= CBC;
  ((PState *)(self->state))->flags &= ~DECRYPT;
  return XK_SUCCESS;
}

static xkern_return_t	
TCRYPT(parameter_DCBC)( self, ptr, fields, line, arg )
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS " DCBC mode");
  ((PState *)(self->state))->flags |= CBC;
  ((PState *)(self->state))->flags |= DECRYPT;
  return XK_SUCCESS;
}

static xkern_return_t	
TCRYPT(parameter_ECFB)( self, ptr, fields, line, arg )
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS " ECFB mode");
  ((PState *)(self->state))->flags |= CFB;
  ((PState *)(self->state))->flags &= ~DECRYPT;
  return XK_SUCCESS;
}

static xkern_return_t	
TCRYPT(parameter_DCFB)( self, ptr, fields, line, arg )
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS " DCFB mode");
  ((PState *)(self->state))->flags |= CFB;
  ((PState *)(self->state))->flags |= DECRYPT;
  return XK_SUCCESS;
}

static xkern_return_t	
TCRYPT(parameter_EOFB)( self, ptr, fields, line, arg )
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS " EOFB mode");
  ((PState *)(self->state))->flags |= OFB;
  ((PState *)(self->state))->flags &= ~DECRYPT;
  return XK_SUCCESS;
}

static xkern_return_t	
TCRYPT(parameter_DOFB)( self, ptr, fields, line, arg )
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS " DOFB mode");
  ((PState *)(self->state))->flags |= OFB;
  ((PState *)(self->state))->flags |= DECRYPT;
  return XK_SUCCESS;
}



void
TCRYPT(_init)( self )
     XObj self;
{
    PState	*ps;
    XObj	llp;

    xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "_init entered");
    xAssert(xIsProtocol(self));
    
    if ( !self->numdown )
      {
	xError(TCRYPTS "_init: " TCRYPTS " need at least one lower protocol object");
	/* return; */
      }

    if (self->numdown == 1)
      xTrace0(TCRYPTP, TR_MAJOR_EVENTS, TCRYPTS " init. Mode uses no transport protocol");

    if ( ! xIsProtocol(llp = xGetDown(self, TRANSPORT_PROTOCOL)) )
      {
	xError(TCRYPTS "_init: " TCRYPTS " cannot get lower protocol object");
	/* return; */
      }

    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->actMap = mapCreate(CRYPT_ACTIVE_MAP_SZ, sizeof(ActiveId));
    ps->pasMap = mapCreate(CRYPT_PASSIVE_MAP_SZ, sizeof(PassiveId));
    ps->flags = 0;  /* Flag &1 turns on IPpseudoheader length fixup.
		       Flag &2 turns on Ivec-from-Key for all sessions.
		       There are many more flags. */
    ps->flags &= ~DECRYPT;
    ps->interleave = 0;

    self->control = TCRYPT(ControlProtl);
    self->open = TCRYPT(Open);
    self->openenable = TCRYPT(OpenEnable);
    self->opendisable = TCRYPT(OpenDisable);
    self->opendone = TCRYPT(OpenDone);
    self->close = TCRYPT(Close);  
    self->demux = TCRYPT(Demux);

    findXObjRomOpts(self, rmopts, sizeof(rmopts)/sizeof(rmopts[0]), 0);
    xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "_init exiting");
}

static int
xCopyParts(q, qbuf, p, lls)
     Part **q;
     char **qbuf;
     Part *p;
     XObj lls;
{
  char *ebuf;
  int	len = 500;

  ebuf = xMalloc(500);
  if (lls != ERR_XOBJ) {
    if (xControl(lls, GETPARTICIPANTS, ebuf, len)<0)
      return 0;
  }
  else
    partExternalize(p, ebuf, &len);
  *q = (Part *)xMalloc(partExtLen(ebuf)*sizeof(Part));
  partInternalize(*q, ebuf);
  *qbuf = ebuf;
  return partLen(((Part *)(*q)));
}

static void
xSimpleCopyParts(q, p)
     Part **q;
     Part *p;
{
  int  pl =  partLen(p);

  *q = (Part *)xMalloc(pl*sizeof(Part));
  partInit(*q, pl);
  while (pl--) { (*q)[pl] = p[pl]; }
}


static XObj
TCRYPT(Open)( self, hlpRcv, hlpType, part )
    XObj	self, hlpRcv, hlpType;
    Part 	*part;
{
    XObj 	TCRYPT(Sessn);
    XObj	lls = ERR_XOBJ;
    PState	*pstate;

    pstate = (PState *) self->state;
    
    if ( part == 0 || partLen(part) < 1 || partLen(part) > 3 ) {
	xTrace0(TCRYPTP, TR_SOFT_ERRORS, TCRYPTS "Open -- bad participants");
	return ERR_XOBJ;
    }

    if (self->numdown>1) {
      xTrace4(TCRYPTP,TR_EVENTS,TCRYPTS "Open calling xOpen with args %x %x %x %x",
	      self,hlpType,xGetDown(self,TRANSPORT_PROTOCOL),part);

      lls = xOpen(self, hlpType, xGetDown(self, TRANSPORT_PROTOCOL), part);
    }
      else
	lls = cryptipseckmOpen(part, self, 0);

    if ( lls == ERR_XOBJ ) {
        xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "Open: could not open lls");
        return ERR_XOBJ;
      }

    if (mapResolve(pstate->actMap, &lls, &TCRYPT(Sessn)) == XK_SUCCESS ) {
        xTrace0(TCRYPTP, TR_MAJOR_EVENTS, "found an existing one");
        xClose(lls);
    } else {
        xTrace0(TCRYPTP, TR_MAJOR_EVENTS, "creating a new one");
        if ( (TCRYPT(Sessn) = TCRYPT(CreateSessn)(self, hlpRcv, hlpType, lls))
	    == ERR_XOBJ ) {
            xClose(lls);
        }
    }

    return (TCRYPT(Sessn));
}

static XObj
cryptipseckmOpen(part, self, kmpi)
     Part *part; XObj self; int kmpi;
{
	Part p[1];
	int  lsaid;
	IPhost myhost, spihost;
	XObj kms1;

     	if (partStackTopByteLen(part[LOCAL_PART]) != sizeof(IPhost))
     	{
 		xTrace0(TCRYPTP, TR_MAJOR_EVENTS,
		    	"open error: ipsecmode: local host is not specified or not an IPhost");
		return ERR_XOBJ;
     	}
     	myhost=*(IPhost *)partPop(part[LOCAL_PART]);
     	partPush(part[LOCAL_PART], (char *)&myhost, sizeof(IPhost));
	spihost = *((IPhost *)partPop(part[SPI_PART]));
	lsaid = *((XSaid *)partPop(part[SPI_PART]));

	xTrace0(TCRYPTP, TR_DETAILED, "CreateSessn: one keym sessn using lsaid");
     	partInit(p,1);
     	partPush(p[0], (char *)&lsaid,sizeof(XSaid));	
     	partPush(p[0], (char *)&spihost, sizeof(IPhost));
	kms1 = xOpen(self, self, xGetDown(self, kmpi), p);
	if (kms1 == ERR_XOBJ) {
		 xTrace0(TCRYPTP, TR_ERRORS, "CreateSessn: keymgr open error");
	        return ERR_XOBJ;
	}
	return kms1;
}


static XObj
TCRYPT(CreateSessn)( self, hlpRcv, hlpType, lls)
    XObj	self, hlpRcv, hlpType, lls;
{
    Part	*part;
    XObj	s;
    SState	*ss;
    PState	*ps = (PState *)self->state;
    XObj        SessnArray[3];
    XObj        kms1;
    XObj        kms2=NULL;
    Part	part1[MOST_PARTS];
    Part	part2[MOST_PARTS];
    int		down_count = 1;
    char	*b = 0;
    int         kmpi = KEY_MGR;

    xTrace4(TCRYPTP, TR_EVENTS, TCRYPTS "CreateSessn entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, lls);

    ss = X_NEW(SState);

    if (self->numdown==1) kmpi = 0;
    else {
      xCopyParts(&part, &b, 0, lls);

      if (partLen(part) < 2) {
	xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "CreateSessn: too few participants");
	xFree((char *) ss);
	return ERR_XOBJ;
      }
    }

    if (!(ps->flags & USE_IPSEC))
    {
        if (self->numdown == 1)
	  xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "CreateSessn: no transport protocol, no ipsec, probably will fail");
    	if (!(ps->flags & ALL_PARTS)) {
    	/* open two key manager sessions */
    	   partInit(part1,1);
   	   part1[0] = part[REMOTE_PART];
   	   part1[0].len =1; /* major hack */
   	   kms1 = xOpen(self, self, xGetDown(self,kmpi), part1);
	      if (kms1 == ERR_XOBJ) { 
		xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "CreateSessn: could not open key manager");
		xFree((char *) ss);
		return ERR_XOBJ;
	      }

	      partInit(part2,1);
	      part2[0] = part[LOCAL_PART];
	      part2[0].len =1; /* major hack */
	      kms2 = xOpen(self, self, xGetDown(self,kmpi), part2);
 	     if (kms2 == ERR_XOBJ) { 
		xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "CreateSessn: could not open key manager");
		xClose(kms1);
		xFree((char *) ss);
		return ERR_XOBJ;
 	     }
       }
       else {
	      kms1 = xOpen(self, self, xGetDown(self,kmpi), part);
	      if (kms1 == ERR_XOBJ) { 
		xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "CreateSessn: could not open key manager");
		if (b) xFree((char *)part);
		xFree((char *) ss);
		return ERR_XOBJ;
	      }
       }
    }
    else
      if ( self->numdown > 1) 
	  if ((kms1 = cryptipseckmOpen(part, self, kmpi)) == ERR_XOBJ) {
	    if (b) xFree((char *)b);
	    return ERR_XOBJ;
	  }
    if (b) xFree((char *)b);
    /* the first down session might be a dummy */
    SessnArray[0]= self->numdown>1 ? lls : ERR_XOBJ;
    /* the key manager will always be at the same place */
    SessnArray[1]= (self->numdown==1 && 
		    (ps->flags&USE_IPSEC||ps->flags & ALL_PARTS))
                     ? lls : kms1;
    down_count = 2;
    if ((!(ps->flags & USE_IPSEC)) && (!(ps->flags & ALL_PARTS)))
      { 
	ps = (PState *)self->state;
	SessnArray[2]=kms2; 
	ps = (PState *)self->state;
	down_count++;
	ps = (PState *)self->state;
      }
    s = xCreateSessn(TCRYPT(SessnInit), hlpRcv, hlpType, self, down_count, SessnArray);
    s->binding = (Bind) mapBind(ps->actMap, (char *)&lls, s);
    
    if ( s->binding == ERR_BIND ) {
      xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "CreateSessn: mapBind error");
      xClose(kms1);
      if (!(ps->flags & ALL_PARTS)) xClose(kms2);
      xDestroy(s);
      xFree((char *) ss);
      return ERR_XOBJ;
    }

    xTrace1(TCRYPTP, TR_FULL_TRACE, TCRYPTS "CreateSessn: X_NEW(SState) returned %x", ss);

    ss->method = 0;
    ss->flags = 0;
	/* temporarily init method & cryptkey here */
    ss->method = 1;

    s->state = (VOID *)ss;
    xTrace1(TCRYPTP, TR_MORE_EVENTS, TCRYPTS "CreateSessn returning %x",s);
    return s;
}


static xkern_return_t
TCRYPT(OpenEnable)(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;
    Part	*qart;

    if (ps->flags & ALL_PARTS) {
      if (ps->flags & ALL_PARTS) xSimpleCopyParts(&qart, part);
    }

    xTrace4(TCRYPTP, TR_EVENTS, TCRYPTS "OpenEnable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    if ( defaultOpenEnable(ps->pasMap, hlpRcv, hlpType, 
	(VOID *)&hlpType) == XK_FAILURE ) {
        xTrace0(TCRYPTP, TR_MORE_EVENTS, TCRYPTS "OpenEnable defaultOpenEnable failed ");
        return XK_FAILURE;
    }
    
    val = xOpenEnable(self, hlpType, xGetDown(self,TRANSPORT_PROTOCOL), part);
    if ( val == XK_FAILURE ) {
        xTrace0(TCRYPTP, TR_MORE_EVENTS, TCRYPTS "OpenEnable xOpenEnable failed ");
        return XK_FAILURE;
    }
    
    if (ps->flags & ALL_PARTS) {
      val = xOpenEnable(self, self, xGetDown(self,KEY_MGR), qart);
      if (val == XK_FAILURE) { 
/* xOpenDisable(self, hlpType, xGetDown(self, TRANSPORT_PROTOCOL), part); */
	xFree((char *)qart);
	xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "CreateSessn: could not open key manager");
	return XK_FAILURE;
      }
    }


    xTrace1(TCRYPTP, TR_MORE_EVENTS, TCRYPTS "OpenEnable returning %x", val);
    return val;
} 


static xkern_return_t
TCRYPT(OpenDisable)(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;



    xTrace4(TCRYPTP, TR_EVENTS, TCRYPTS "OpenDisable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    val = defaultOpenDisable(ps->pasMap, hlpRcv, hlpType, (void*)&hlpType);

    xTrace1(TCRYPTP, TR_MORE_EVENTS, TCRYPTS "OpenDisable returning %x", val);
    return val;
}

static xkern_return_t
TCRYPT(OpenDone)(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    XObj        s;
    Enable	*e;
    PState      *pstate;
   
    xTrace0(TCRYPTP, TR_MAJOR_EVENTS, "Entered " TCRYPTS "OpenDone");

    if ( self == hlpType ) {
        xTrace0(TCRYPTP, TR_ERRORS, "self == hlpType in " TCRYPTS "OpenDone");
        return XK_FAILURE;
    }
    pstate = (PState *)self->state;

    /*
     * check for open enables
     */
    if ( mapResolve(pstate->pasMap, &hlpType, &e) == XK_FAILURE ) {
        /*
         * This shouldn't happen
         */
        xTrace0(TCRYPTP, TR_ERRORS,
                TCRYPTS "OpenDone: Couldn't find hlp for incoming session");
        return XK_FAILURE;
    }

    xDuplicate(lls);
    if ( (s = TCRYPT(CreateSessn)(self, e->hlpRcv, e->hlpType, lls))
                == ERR_XOBJ ) {
        xClose(lls);
        return XK_FAILURE;
    }

    xTrace0(TCRYPTP, TR_EVENTS,
            TCRYPTS "OpenDone: Passively opened session successfully created");
    return xOpenDone(e->hlpRcv, s, self);
}

static xkern_return_t
TCRYPT(Demux)( self, lls, msg )
    XObj	self, lls;
    Msg		*msg;
{
  PState	*ps = (PState *)self->state;
  XObj 	s = 0;
  xkern_return_t val = XK_FAILURE;
  
  xTrace3(TCRYPTP, TR_EVENTS, TCRYPTS "Demux entered: args %x %x %x",
	  self, lls, msg);

  xIfTrace(TCRYPTP, TR_DETAILED) msgShow(msg);

  if ( mapResolve(ps->actMap, &lls, &s) == XK_FAILURE ) { 
	xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "Demux: map resolve fails");
  	return XK_FAILURE;
  }

  if ( xIsSession(s) )  val = xPop(s, lls, msg, 0);
  else val = XK_FAILURE;

  xIfTrace(TCRYPTP, TR_FUNCTIONAL_TRACE)
  {
  if (val == XK_SUCCESS)
  xTrace0(TCRYPTP, TR_FUNCTIONAL_TRACE, TCRYPTS "Demux returning XK_SUCCESS");
  else if (val == XK_FAILURE)
  xTrace0(TCRYPTP, TR_FUNCTIONAL_TRACE, TCRYPTS "Demux returning XK_FAILURE");
  else
  xTrace1(TCRYPTP, TR_FUNCTIONAL_TRACE, TCRYPTS "Demux returning %x", val);
  }

  return val;
}


static xkern_return_t
TCRYPT(Close)( s )
    XObj	s;
{   SState *ss;
    PState	*ps = (PState *)xMyProtl(s)->state;
    xkern_return_t res;

    xTrace1(TCRYPTP, TR_MAJOR_EVENTS, TCRYPTS "Close: closing session %x", s);
    xAssert( xIsSession( s ) );
    xAssert( s->rcnt == 0 );
    res = mapRemoveBinding( ps->actMap, s->binding );
    xAssert( res != XK_FAILURE );
    if (s->myprotl->numdown > 1) xClose(xGetDown(s,TRANSPORT_PROTOCOL));
    if (s->numdown > 1) xClose(xGetDown(s, KEY_MGR_OUT));
    if (s->numdown > 2) xClose(xGetDown(s, KEY_MGR_IN));

    /* zero out any cryptographically important state */
    ss = s->state;
    ss->method = 0; ss->flags = 0;

    xDestroy( s );
    xTrace0(TCRYPTP, TR_DETAILED, TCRYPTS "Close returning");
    return XK_SUCCESS;
}


static
long TCRYPT(copy)(to,from,len,ignore)
    VOID *to; char *from; VOID *ignore; long len;
{
  xTrace4(TCRYPTP, TR_FULL_TRACE, TCRYPTS "copy args: %x %x %x %x",
	  to,from,len,ignore);

    bcopy(from,to,len);
  return len;
}

static
void TCRYPT(Store)(src, dst, len, arg)
    VOID *src;
    char *dst;
    long len;
    VOID *arg;
{

    bcopy((char *)src,dst,len);
}

static xmsg_handle_t
TCRYPT(Push)( s, msg )
    XObj	s;
    Msg 	*msg;
{
    long len, lm8, clen, i, tmp, clentotal, left, tot, fragct, ckz=0;  
    Msg cmsg; /* the ciphertext message */
    char *cbuf, *cbufreal;
    int crval;
    u_char cryptkey[MAXKEYSZ+1];
    XObj        kms = xGetDown(s, KEY_MGR_OUT);
    PState      *ps = (PState *)xMyProtl(s)->state;
    long     pflags = ps->flags;
    SState      *ss = s->state;
    long     sflags = ss->flags;
    long    psflags = pflags | sflags;

    xTrace4(TCRYPTP, TR_EVENTS, TCRYPTS
	    "Push: session %x  msg %x  sflg %x  pflg %x", s, msg, sflags, pflags);
    xIfTrace(TCRYPTP, TR_EVENTS) 
    { print_crypt_session_state(TCRYPTS "Push: ",ss,pflags);
      printf("\nideaPush: show the plaintext message\n");
      msgShow(msg);
    }

    /* catch one terrible flag combination that would ship the key */
    if ((psflags & (SEND_IV | SEND_IV_ONCE)) &&
	(psflags & IV_FROM_KEY) && !(psflags & CRYPT_IV))
      {	xTrace0( TCRYPTP, TR_ERRORS,
	    "ERROR: " TCRYPTS "Push() -- Illegal flag combination -- Dropping message");
	return XK_FAILURE;  }


  { int keysize;
    int nokey;

    nokey = 0;
    for (i=0; i <= MAXKEYSZ; i++) cryptkey[i] = 0;
    keysize = xControl(kms, KM_RESOLVE, (char *)cryptkey, MAXKEYSZ);   
    if ((keysize < MINKEYSZ) || (keysize > MAXKEYSZ))   nokey = 1;
    for ( i = 0; i < keysize; i++ )
      if ( cryptkey[i] != 0 ) break;
    if ( i >= keysize )  nokey = 1;
    if ( nokey )
    {
	/* We did not find a valid encryption key. */
	/* Don't send cleartext, so simply drop the message */

	xTrace0( TCRYPTP, TR_ERRORS,
	    "ERROR: " TCRYPTS "Push() -- No encryption key found -- Dropping message");
	return XK_FAILURE;
    }
  }

    xIfTrace(TCRYPTP, TR_FULL_TRACE)
    { printf("cryptkey %x: %x %x %x %x %x %x %x %x   ", cryptkey,
	   cryptkey[0], cryptkey[1], cryptkey[2], cryptkey[3],
	   cryptkey[4], cryptkey[5], cryptkey[6], cryptkey[7]);
    };

    /* Encrypt here for now.  This should be broken out later. */
    len = msgLen(msg);
    { lm8 = len%8;  /* Using pad-to-multiple-of-8 hack from Sun DES program. */
      clen = len + 8 - lm8; /* ciphertext grows 1-8 bytes, to next multiple of 8 */ 
      clen += 8;  /* add another 8 byte for idea, coz idea will add the len of msg to the last 8 byte */
    }

    clentotal = clen;

    /* I discard the message attribute field on the way down.  "Feature." */
    msgConstructAllocate(&cmsg, clentotal, &cbuf);
    cbufreal = cbuf;
    left = len;
    if ((psflags & ROTATE_BYTE) && len)
    { msgPop(msg, TCRYPT(copy),cbuf,1,0); tmp = cbuf[0]; left--; }
    msgPop(msg, TCRYPT(copy), cbuf, left, 0);  /* serialize message */

    if (len<clen-8)  /* If padding required. */
    { for (i=left;i<clen-1;i++) *(cbuf+i)=0;  /* append 7-lm8 zeros, then lm8 */
      if (psflags & RFC_1829)  
		*(cbuf+clen-1-8) = 7-lm8;  /* Why? */
	else
		*(cbuf+clen-1-8) = lm8; }; /* put len%8 */
    /* the last 8 bytes is left for idea, so cbuf+clen-9 rather than cbuf+clen-1 */
    if ((psflags & ROTATE_BYTE) && len)
    { /* If padding, we did an extra 0, and pad must be moved one space earlier */
      if (len<clen-8) *(cbuf+clen-2-8) = 7-lm8;
      *(cbuf+clen-1-8) = tmp;  /* put rotated byte after padlength byte */
      xTrace2(TCRYPTP, TR_FULL_TRACE, " Pad length %u , Payload type %u",
		*(cbuf+clen-2-8), *(cbuf+clen -1-8));
      tmp = 0;  /* clear tmp */ }
    /* how to make sure that compiler doesn't discard tmp=0 as irrelevant? $$$$ */

    xIfTrace(TCRYPTP, TR_FULL_TRACE)
    { printf("ptx ");
      for (i=0;i<clentotal && i<40;i++) printf("%02x",0xff & *(cbufreal+i)); printf("\n");
      if (clentotal>=40) { printf("ptx ");
			   for (i=(clentotal<80)?40:(clentotal-40);i<clentotal;i++)
			     printf("%02x",0xff & *(cbufreal+i));        
                           printf("\n");
                         };
    }

    crval = xk_idea_crypt(cryptkey, cbuf, clen, &ps->flags, ps->interleave);
    xTrace2(TCRYPTP, TR_FULL_TRACE, TCRYPTS " mode %x %x",
            ps->flags, ps->interleave);

    bzero((char *)cryptkey, sizeof(cryptkey)); ckz++;
    xIfTrace(TCRYPTP, TR_FULL_TRACE)
    { printf("ctx ");
      for (i=0;i<clentotal && i<40;i++) printf("%02x",0xff & *(cbufreal+i)); printf("\n");
      if (clentotal>=40) { printf("ctx ");
			   for (i=(clentotal<80)?40:(clentotal-40);i<clentotal;i++)
			     printf("%02x",0xff & *(cbufreal+i));         printf("\n");
                         };
    }

    if (crval) {
       xTrace1(TCRYPTP, TR_ERRORS, TCRYPTS "Push: Encryption barfed; error %D", crval);
    }
    else { ss->flags &= ~SEND_IV_ONCE;
	   if (s->myprotl->numdown>1) xPush(xGetDown(s, TRANSPORT_PROTOCOL), &cmsg);
	   else msgAssign(msg, &cmsg);
	   /* leaf mode: replace the original msg with the encrypted one */
    }

    xIfTrace(TCRYPTP, TR_EVENTS)  
    {  xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS " show encryptted message:\n");
       msgShow(&cmsg);
    }
    msgDestroy(&cmsg);

    if (!ckz) bzero((char *)cryptkey, sizeof(cryptkey));
    xIfTrace(TCRYPTP, TR_FULL_TRACE) print_crypt_session_state(TCRYPTS "Push: ",ss,pflags);
    xTrace0(TCRYPTP, TR_DETAILED, TCRYPTS "Push exiting");
    return crval ? XK_FAILURE : XMSG_NULL_HANDLE;
}


static xkern_return_t
TCRYPT(Pop)( s, lls, m, hdr)
    XObj	s, lls;
    Msg 	*m;
    void        *hdr;
{
    long len,orglen,dlen,dbuflen,fragct; 
    char* dbuf, *dbufreal, *dbuftmp;
    int crval=0,i,nomsg=0,ckz=0;
    int lm8;
    u_char cryptkey[MAXKEYSZ];
    XObj   kms;
    PState      *ps = (PState *)xMyProtl(s)->state;
    SState      *ss = s->state;
    long     pflags = ps->flags;
    long     sflags = ss->flags;
    long    psflags = pflags | sflags;

    if (s->numdown > 2) 
	kms = xGetDown(s,KEY_MGR_IN);
    else
	kms = xGetDown(s,KEY_MGR);
	
    xTrace4(TCRYPTP, TR_EVENTS, TCRYPTS "Pop: args %x %x %x %x", s, lls, m, hdr);
    xIfTrace(TCRYPTP, TR_FULL_TRACE) print_crypt_session_state(TCRYPTS "Pop: ",ss,pflags);
    xIfTrace(TCRYPTP, TR_EVENTS)  
    {  xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "show encryptted message got:\n");
       msgShow(m);
    }

 {  int keysize;
    keysize = xControl(kms, KM_RESOLVE, (char *)cryptkey, MAXKEYSZ);
    cryptkey[keysize] = 0;
    if ((keysize < MINKEYSZ) || (keysize > MAXKEYSZ))
    { xTrace0(TCRYPTP, TR_ERRORS, "key lookup failed for inbound message, dropping it");
      return XK_FAILURE; 
    }
  }

    xIfTrace(TCRYPTP, TR_MAJOR_EVENTS)
    { printf("cryptkey %x: %x %x %x %x %x %x %x %x   ", cryptkey,
	   cryptkey[0], cryptkey[1], cryptkey[2], cryptkey[3],
	   cryptkey[4], cryptkey[5], cryptkey[6], cryptkey[7]); 
    }

    /* Decrypt here for now.  This should be broken out later. */
    orglen = len = msgLen(m);
    if (len%8)
    { xTrace1(TCRYPTP, TR_ERRORS,
	      "cryptPop received bad message, length %D not a multiple of 8",len);
      return XK_FAILURE; 
    }
    dbuflen = len+16;  /* leave 8 bytes at front if I need to undo ROTATE_BYTE */
    dbufreal = xMalloc(dbuflen);
    dbuftmp = dbuf = dbufreal+8;
    msgPop(m, TCRYPT(copy), dbuftmp, len, 0);  /* serialize message */

    xIfTrace(TCRYPTP, TR_FULL_TRACE)
    { printf("ctx "); 
      for (i=0;i<len && i<40;i++) printf("%02x",0xff & *(dbuf+i)); printf("\n");
      if (len>=40) 
      { printf("ctx ");
        for (i=(len<80)?40:(len-40);i<len;i++) printf("%02x",0xff & *(dbuf+i)); printf("\n");
      };
    }

    /* in Pop, reverse the decrypt/encrypt flag of the Push */
    if (ps->flags & DECRYPT) ps->flags &= ~DECRYPT; else  ps->flags |= DECRYPT;
    crval = xk_idea_crypt(cryptkey,dbuf,len, &ps->flags, ps->interleave);
    if (ps->flags & DECRYPT) ps->flags &= ~DECRYPT; else  ps->flags |= DECRYPT;
    /* After Pop, restore the decrypt/encrypt flag */
    
    xTrace2(TCRYPTP, TR_FULL_TRACE, TCRYPTS " mode %x %x", ps->flags, ps->interleave);

    bzero((char *)cryptkey, sizeof(cryptkey)); ckz++;
    xIfTrace(TCRYPTP, TR_FULL_TRACE)
    { printf("ptx "); for (i=0;i<len && i<40;i++) printf("%02x",0xff & *(dbuf+i)); printf("\n");
      if (len>=40) 
      { printf("ptx ");
        for (i=(len<80)?40:(len-40);i<len;i++) printf("%02x",0xff & *(dbuf+i)); printf("\n");
      };
    }

    if (crval)
    { xTrace1(TCRYPTP, TR_ERRORS, TCRYPTS "Pop: Decryption barfed; error %D", crval);
      goto free; 
    }

    len -= 8;  /* because I add another 8 bytes for IDEA */

    /* unrotate byte */
    if ((psflags & ROTATE_BYTE) && len) { dbuf--; *dbuf = *(dbuf+len); }
 
    if (psflags & RFC_1829)  
/*
		if (psflags & ROTATE_BYTE) 
			lm8 = 7- (*(dbuf+len-2));
		else
*/
			lm8 = 7 - (*(dbuf+len-1));
	else
		lm8 = (*(dbuf+len-1));

    /* inspect last byte, should be 0-7 */
    if (lm8>7)
    { xTrace1(TCRYPTP, TR_ERRORS, TCRYPTS "Pop: decryption failed; last byte (%D) >7", lm8);
      crval = 1;  
      goto free; 
    }
    dlen = len - 8 + lm8; /*  remove (8-lastbyte) from end (should be 0s, except for last byte) */
    if (!(psflags & NO_PADDING_CHECK))
    { for (i=dlen; i<len-1; i++)
       if (*(dbuf+i))
         { xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "Pop: decryption failed; pad!=0");
	 crval = 1; goto free; } }
    /*
    for (i=dlen; i<len-1; i++)
      if (*(dbuf+i))
      { xTrace0(TCRYPTP, TR_ERRORS, TCRYPTS "Pop: decryption failed; pad!=0");
	crval = 1; goto free; 
      }
    */

    /* This way, message attributes are preserved on the way up! */
    msgPush(m, TCRYPT(Store),dbuf,dlen, 0);

    /* Fixup length in IP pseudoheader? */
    if (psflags & IP_PSEUDOHDR_HACK)
    {  IPpseudoHdr *ipphdr = msgGetAttr( m, 0);
       if (ipphdr)
       {  ipphdr->len = htons(ntohs(ipphdr->len) - (orglen - dlen));
	  xTrace4( TCRYPTP, TR_FULL_TRACE, TCRYPTS "Pop:  Adjusting IP pseudoheader length;\n                   ipphdr %x  phlen after %d  orglen %d  dlen %d",
		  ipphdr, ntohs(ipphdr->len), orglen, dlen);  
       }
       else 
       { xTrace0( TCRYPTP, TR_FULL_TRACE,
	 TCRYPTS "Pop:  Skipping length adjustment for IP pseudoheader - msg attr is 0."); 
       };
     }

free:
    if (!ckz) bzero((char *)cryptkey, sizeof(cryptkey));
    bzero(dbufreal,dbuflen);  /* now that we're not doing timing benchmarks, zero cleartext */
    xFree(dbufreal);  /* $$$$ Should zero out cleartext. */

    xIfTrace(TCRYPTP, TR_FULL_TRACE) print_crypt_session_state(TCRYPTS "Pop: ",ss,pflags);
    if (nomsg) return XK_SUCCESS;
    /* leaf mode(s->up == 0), donot do xDemux(s,m) */
    if (!s->up) return crval ? XK_FAILURE: XK_SUCCESS;

    xIfTrace(TCRYPTP, TR_EVENTS)  
    {  xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "show decryptted message got:\n");
       msgShow(m);
    }
    /* not leaf mode, do xDemux(s,m) */
    return crval ?  XK_FAILURE : xDemux(s, m);
} 


static int
TCRYPT(ControlSessn)(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsSession(s));
    switch (op) {

      case GETMAXPACKET:
      case GETOPTPACKET:
	{ 
	    int max;
	    if (len < sizeof(max) || s->myprotl->numdown < 2) return 0; 
	    if (xControl(xGetDown(s, TRANSPORT_PROTOCOL), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
#ifdef PRESERVE_LENGTH
	    if (max<8) max=0;
#else
 	    max &= (~7); max--; if (max<0) return 0;
#endif
	    *(int *)buf = max;
	    return(sizeof(max));
	}

      case CRYPT_ISPRIVATE: return 1;

      /* redirect to key manager session for incoming packets */
      case CRYPT_IN_RESOLVE: 
	return xControl(xGetDown(s, KEY_MGR_IN), KM_RESOLVE, buf, len);
      case CRYPT_IN_SET: 
	return xControl(xGetDown(s, KEY_MGR_IN), KM_SET, buf, len);
      case CRYPT_IN_INVALIDATE: 
	return xControl(xGetDown(s, KEY_MGR_IN), KM_INVALIDATE, buf, len);
      case CRYPT_IN_ISEMPTY: 
	return xControl(xGetDown(s, KEY_MGR_IN), KM_ISEMPTY, buf, len);
      case CRYPT_IN_ISVALID: 
	return xControl(xGetDown(s, KEY_MGR_IN), KM_ISVALID, buf, len);
      case CRYPT_IN_ISINVALID: 
	return xControl(xGetDown(s, KEY_MGR_IN), KM_ISINVALID, buf, len);
      case CRYPT_IN_KEYSIZE: 
	return xControl(xGetDown(s, KEY_MGR_IN), KM_KEYSIZE, buf, len);
      case CRYPT_IN_HOSTSIZE: 
	return xControl(xGetDown(s, KEY_MGR_IN), KM_HOSTSIZE, buf, len);

      /* redirect to key manager session for outgoing packets */
      case CRYPT_OUT_RESOLVE: 
	return xControl(xGetDown(s, KEY_MGR_OUT), KM_RESOLVE, buf, len);
      case CRYPT_OUT_SET: 
	return xControl(xGetDown(s, KEY_MGR_OUT), KM_SET, buf, len);
      case CRYPT_OUT_INVALIDATE: 
	return xControl(xGetDown(s, KEY_MGR_OUT), KM_INVALIDATE, buf, len);
      case CRYPT_OUT_ISEMPTY: 
	return xControl(xGetDown(s, KEY_MGR_OUT), KM_ISEMPTY, buf, len);
      case CRYPT_OUT_ISVALID: 
	return xControl(xGetDown(s, KEY_MGR_OUT), KM_ISVALID, buf, len);
      case CRYPT_OUT_ISINVALID: 
	return xControl(xGetDown(s, KEY_MGR_OUT), KM_ISINVALID, buf, len);
      case CRYPT_OUT_KEYSIZE: 
	return xControl(xGetDown(s, KEY_MGR_OUT), KM_KEYSIZE, buf, len);
      case CRYPT_OUT_HOSTSIZE: 
	return xControl(xGetDown(s, KEY_MGR_OUT), KM_HOSTSIZE, buf, len);

      case CRYPT_IVEC_FROM_KEY:
	{   ((SState *)(s->state))->flags |= IV_FROM_KEY;
	    xTrace0(TCRYPTP, TR_EVENTS,
	      TCRYPTS "ControlSessn: Turning on Ivec-from-Key flag.");

            printf("\nCRYPT_IVEC_FROM_KEY\n"); exit(-1);

	    return 0;  };
      case CRYPT_READ_FLAGS:
	{   long flags = ((SState *)(s->state))->flags;
	    if (len < sizeof(long)) return -1;
	    *(long *)buf = flags;
	    xTrace1(TCRYPTP, TR_EVENTS,
	      TCRYPTS "ControlSessn: Reading flags: %x.",flags);
	    return sizeof(long);  }
      case CRYPT_SET_FLAGS:
	{   long oflags = ((SState *)(s->state))->flags,nflags;
	    if (len < sizeof(long)) return -1;
	    nflags = ((SState *)(s->state))->flags = oflags | *(long *)buf;
	    xTrace3(TCRYPTP, TR_EVENTS,
	      TCRYPTS "ControlSessn: setting flags: old %x, setting %x, new %x.",
		    oflags,*(long *)buf,nflags);
	    return 0;  }
      case CRYPT_CLEAR_FLAGS:
	{   long oflags = ((SState *)(s->state))->flags,nflags;
	    if (len < sizeof(long)) return -1;
	    nflags = ((SState *)(s->state))->flags = oflags &~ *(long *)buf;
	    xTrace3(TCRYPTP, TR_EVENTS,
	      TCRYPTS "ControlSessn: clearing flags: old %x, clearing %x, new %x.",
		    oflags,*(long *)buf,nflags);
	    return 0;  }
      case CRYPT_READ_IV_DOWN:
	{   xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "ControlSessn: reading IV_DOWN: ");
	    xIfTrace(TCRYPTP, TR_DETAILED) hexbprint(((SState *)(s->state))->iv_down, 8);

            printf("\nCRYPT_READ_IV_DOWN\n"); exit(-1);

	    if (len>=8) { bcopy(((SState *)(s->state))->iv_down, buf, 8); return 8; }
	    else return -1; }
      case CRYPT_SET_IV_DOWN:
	{   xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "ControlSessn: setting IV_DOWN: ");

            printf("\nCRYPT_SET_IV_DOWN\n"); exit(-1);

	    xIfTrace(TCRYPTP, TR_DETAILED)
	      { printf("was "); hexbprint(((SState *)(s->state))->iv_down, 8);
		printf("new "); if (buf) hexbprint(buf, 8); };
	    if (len>=8) { bcopy(buf, ((SState *)(s->state))->iv_down, 8); return 0; }
	    else return -1; }
      case CRYPT_READ_IV_UP:
	{   xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "ControlSessn: reading IV_UP: ");
	    xIfTrace(TCRYPTP, TR_DETAILED) hexbprint(((SState *)(s->state))->iv_up, 8);

            printf("\nCRYPT_READ_IV_UP\n"); exit(-1);

	    if (len>=8) { bcopy(((SState *)(s->state))->iv_up, buf, 8); return 8; }
	    else return -1; }
      case CRYPT_SET_IV_UP:
	{   xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "ControlSessn: setting IV_UP: ");

            printf("\nCRYPT_SET_IV_UP\n"); exit(-1);

	    xIfTrace(TCRYPTP, TR_DETAILED)
	      { printf("was "); hexbprint(((SState *)(s->state))->iv_up, 8);
		printf("new "); if (buf) hexbprint(buf, 8); };
	    if (len>=8) { bcopy(buf, ((SState *)(s->state))->iv_up, 8); return 0; }
	    else return -1; }
      case CRYPT_READ_IV_RANDOM:
	{   xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "ControlSessn: reading IV_RANDOM: ");
	    xIfTrace(TCRYPTP, TR_DETAILED) hexbprint(((SState *)(s->state))->iv_random, 16);

            printf("\nCRYPT_READ_IV_RANDOM\n"); exit(-1);

	    if (len>=16) { bcopy(((SState *)(s->state))->iv_random, buf, 16); return 16; }
	    else return -1; }
      case CRYPT_SET_IV_RANDOM:
	{   xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "ControlSessn: setting IV_RANDOM: ");

            printf("\nCRYPT_SET_IV_RANDOM\n"); exit(-1);

	    xIfTrace(TCRYPTP, TR_DETAILED)
	      { printf("was "); hexbprint(((SState *)(s->state))->iv_random, 16);
		printf("new "); if (buf) hexbprint(buf, 16); };
	    if (len>=16) { bcopy(buf, ((SState *)(s->state))->iv_random, 16); return 0; }
	    else return -1; }
      case CRYPT_READ_FRAG_DOWN:
	{   int ct = ((SState *)(s->state))->frag_down_ct;

            printf("\nCRYPT_READ_FRAG_DOWN\n"); exit(-1);

	    xTrace1(TCRYPTP, TR_EVENTS, TCRYPTS "ControlSessn: reading FRAG_DOWN: %d valid  ",ct);
	    xIfTrace(TCRYPTP, TR_DETAILED) hexbprint(((SState *)(s->state))->frag_down, 8);
	    if (len>=ct) { bcopy(((SState *)(s->state))->frag_down, buf, ct); return ct; }
	    else return -1; }
      case CRYPT_SET_FRAG_DOWN:
	{   xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "ControlSessn: setting FRAG_DOWN: ");

            printf("\nCRYPT_SET_FRAG_DOWN\n"); exit(-1);

	    xIfTrace(TCRYPTP, TR_DETAILED)
	      { printf("was %d valid ",((SState *)(s->state))->frag_down_ct);
		hexbprint(((SState *)(s->state))->frag_down, 8);
		printf("new %d valid ", len); if (buf) hexbprint(buf, 8); };
	    if (len<=8) { bzero(((SState *)(s->state))->frag_down, 8);
			  if (len) bcopy(buf, ((SState *)(s->state))->frag_down, len);
			  return 0; }
	    else return -1; }
      case CRYPT_READ_FRAG_UP:
	{   int ct = ((SState *)(s->state))->frag_up_ct;

            printf("\nCRYPT__READ_FRAG_UP\n"); exit(-1);

	    xTrace1(TCRYPTP, TR_EVENTS, TCRYPTS "ControlSessn: reading FRAG_UP: %d valid  ",ct);
	    xIfTrace(TCRYPTP, TR_DETAILED) hexbprint(((SState *)(s->state))->frag_up, 8);
	    if (len>=ct) { bcopy(((SState *)(s->state))->frag_up, buf, ct); return ct; }
	    else return -1; }
      case CRYPT_SET_FRAG_UP:
	{   xTrace0(TCRYPTP, TR_EVENTS, TCRYPTS "ControlSessn: setting FRAG_UP: ");

            printf("\nCRYPT_SET_FRAG_UP\n"); exit(-1);

	    xIfTrace(TCRYPTP, TR_DETAILED)
	      { printf("was %d valid ",((SState *)(s->state))->frag_up_ct);
		hexbprint(((SState *)(s->state))->frag_up, 8);
		printf("new %d valid ", len); if (buf) hexbprint(buf, 8); };
	    if (len<=8) { bzero(((SState *)(s->state))->frag_up, 8);
			  if (len) bcopy(buf, ((SState *)(s->state))->frag_up, len);
			  return 0; }
	    else return -1; }

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((SState *)(s->state))->flags |= IP_PSEUDOHDR_HACK;
	    xTrace0(TCRYPTP, TR_EVENTS,
	      TCRYPTS "ControlSessn: Turning on IPpseudoheader length fixup.");  }
/* **** FALL THROUGH TO DEFAULT CASE, so lower sessions see it too.   **** */
/* **** Code for this control-op should be immediately above default: **** */

      default:
	if (s->myprotl->numdown == 1) return -1;
        if (buf && len) bzero(buf, len);   /* prevent info going down */
	return xControl(xGetDown(s,TRANSPORT_PROTOCOL), op, buf, len);
    }
}


static int
TCRYPT(ControlProtl)( self, op, buf, len )
    XObj	self;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsProtocol(self));
    switch (op) {
      case GETMAXPACKET:
      case GETOPTPACKET:
        {
            int max;
	    if (len < sizeof(max)) return 0; 
	    if (xControl(xGetDown(self, TRANSPORT_PROTOCOL), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
 	    max &= (~7); max--; if (max<0) return 0;
            *(int *)buf = max;
            return(sizeof(max));
	}

      case CRYPT_IVEC_FROM_KEY:
	{   ((PState *)(self->state))->flags |= IV_FROM_KEY;

            printf("\nCRYPT_IVEC_FROM_KEY\n"); exit(-1);

	    xTrace0(TCRYPTP, TR_EVENTS,
	      TCRYPTS "ControlProtl: Turning on Ivec-from-Key flag.");
	    return 0;  };
      case CRYPT_READ_FLAGS:
	{   long flags = ((PState *)(self->state))->flags;
	    if (len < sizeof(long)) return -1;
	    *(long *)buf = flags;
	    xTrace1(TCRYPTP, TR_EVENTS,
	      TCRYPTS "ControlProtl: Reading flags: %x.",flags);
	    return sizeof(long);  }
      case CRYPT_SET_FLAGS:
	{   long oflags = ((PState *)(self->state))->flags,nflags;
	    if (len < sizeof(long)) return -1;
	    nflags = ((PState *)(self->state))->flags = oflags | *(long *)buf;
	    xTrace3(TCRYPTP, TR_EVENTS,
	      TCRYPTS "ControlProtl: setting flags: old %x, setting %x, new %x.",
		    oflags,*(long *)buf,nflags);
	    return 0;  }
      case CRYPT_CLEAR_FLAGS:
	{   long oflags = ((PState *)(self->state))->flags,nflags;
	    if (len < sizeof(long)) return -1;
	    nflags = ((PState *)(self->state))->flags = oflags &~ *(long *)buf;
	    xTrace3(TCRYPTP, TR_EVENTS,
	      TCRYPTS "ControlProtl: clearing flags: old %x, clearing %x, new %x.",
		    oflags,*(long *)buf,nflags);
	    return 0;  }

      /* for composablity crypt cannot zero buf */
      case ETH_REGISTER_ARP:
	{
	  return xControl(xGetDown(self, TRANSPORT_PROTOCOL), op, buf, len);
	}

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((PState *)(self->state))->flags |= IP_PSEUDOHDR_HACK;
	    xTrace0(TCRYPTP, TR_EVENTS,
	      TCRYPTS "ControlProtl: Turning on IPpseudoheader length fixup.");  }
/* **** FALL THROUGH TO DEFAULT CASE, so lower protocols see it too.  **** */
/* **** Code for this control-op should be immediately above default: **** */

      default:
	{
          if (buf && len) bzero(buf, len);   /* prevent info going down */
	  return xControl(xGetDown(self, TRANSPORT_PROTOCOL), op, buf, len);
	}
  }
  
}


static void
TCRYPT(SessnInit)(s)
    XObj s;
{
  s->push = TCRYPT(Push);
  s->pop = TCRYPT(Pop);
  s->close = TCRYPT(Close);
  s->control = TCRYPT(ControlSessn);
}


