/*
 * join_build_auth.c
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 *
 * $Revision: 1.2 $
 * $Date: 1994/05/03 16:34:13 $
 */

#include "xkernel.h"
#include "join_server.h"
#include "join_server_i.h"
#include <krb5/krb5.h>
#include <isode/psap.h>
#include <krb5/kdb.h>
#include <krb5/asn1.h>
#include <com_err.h>
#include <krb5/asn.1/encode.h>
#include "rom.h"

#ifdef __STDC__

static XObj		joinCreateSessn( XObj, XObj, XObj, XObj);
static XObj 		joinCreateTagSessn (XObj, XObj, XObj, XObj, char *, int);
static void		joinSessnInit( XObj );
static void		joinTagSessnInit( XObj );
static int		joinControlProtl( XObj, int, char *, int );
static xkern_return_t	joinDemux( XObj, XObj, Msg * );
static XObj		joinOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	joinOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	joinOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	joinOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t	joinClose( XObj );
static xkern_return_t	joinTagClose( XObj );
static xmsg_handle_t	joinPush( XObj, Msg * );
static xmsg_handle_t	joinTagPush( XObj, Msg * );
static xkern_return_t	joinPop( XObj, XObj, Msg *, void * );
static xkern_return_t	joinTagPop( XObj, XObj, Msg *, void * );
static int		joinControlSessn( XObj, int, char *, int );
static int		joinTagControlSessn( XObj, int, char *, int );
static int 		part2ToId(Part *, char *, int);
static int 		partToId(Part *, char *, int);
static int 		isOrderValid( int *, int);
static void 		processRomFile(XObj);
static long             mycopypop(void *, char *, long, void *);
static void             mycopypush(void *, char *, long, void *);
static int              joindone_proc( XObj, int, char *, int );

#else

static XObj		joinCreateSessn();
static XObj 		joinCreateTagSessn ();
static void		joinSessnInit();
static void		joinTagSessnInit();
static int		joinControlProtl();
static xkern_return_t	joinDemux();
static XObj		joinOpen();
static xkern_return_t	joinOpenEnable();
static xkern_return_t	joinOpenDisable();
static xkern_return_t	joinOpenDone();
static xkern_return_t	joinClose();
static xkern_return_t	joinTagClose();
static xmsg_handle_t	joinPush();
static xmsg_handle_t	joinTagPush();
static xkern_return_t	joinPop();
static xkern_return_t	joinTagPop();
static int		joinControlSessn();
static int		joinTagControlSessn();
static int 		part2ToId();
static int 		partToId();
static int 		isOrderValid( int *, int);
static void 		processRomFile();
static long             mycopypop();
static long             mycopypush();
static int              joindone_proc();

#endif /* __STDC__ */

int tracejoin_serverp = 1;
#define JOINP join_serverp

void
join_server_init( self )
    XObj self;
{
    PState	*ps;
    XObj	llp;
    int 	i;

    xAssert(xIsProtocol(self));
    
    if ( ! xIsProtocol(llp = xGetDown(self, 0)) ) {
	xError("join_init: cannot get lower protocol object");
	return;
    }

    ps = X_NEW(PState);
    bzero((char *)ps,sizeof(ps));
    self->state = (VOID *)ps;
    ps->actMap = mapCreate(JOIN_ACTIVE_MAP_SZ, sizeof(ActiveId));
    ps->tagIdLen = 0;
    ps->pasIdLen = 0;
    ps->numSeg = DEFAULT_NUMSEG;
    for (i=0; i<JOINMAXSEG; i++) {
        ps->order[i] =i;
    } 
    processRomFile(self);


    if (!isOrderValid(ps->order,ps->numSeg)) {
	Kabort("joinControlProtl: invalid order");
    }

    self->control = joinControlProtl;
    self->open = joinOpen;
    self->openenable = joinOpenEnable;
    self->opendisable = joinOpenDisable;
    self->opendone = joinOpenDone;
    self->close = joinClose;  
    self->demux = joinDemux;

}



static XObj
joinOpen( self, hlpRcv, hlpType, part )
    XObj	self, hlpRcv, hlpType;
    Part 	*part;
{
    XObj 	joinSessn;
    XObj 	tagSessn;
    XObj	lls;
    PState	*ps = (PState *)self->state;
    char	idBuf[MAXID];
    char	partBuf[MAXID*2];
    Part	fullPart[MAXPART];
    int		idLen;
    

    if ( part == 0 || partLen(part) < 1 || partLen(part) > 2 ) {
	return ERR_XOBJ;
    }

    idLen = partToId(part,idBuf, MAXID);

    if ((ps->tagIdLen == 0) || 
        (idLen != ps->tagIdLen) ||
        (mapResolve(ps->tagMap,idBuf, &tagSessn) != XK_SUCCESS)) {

	/* new connection required */
        lls = xOpen(self, hlpType, xGetDown(self, 0), part);
        if ( lls == ERR_XOBJ ) {

            return ERR_XOBJ;
        }
        if (xControl(lls, GETPARTICIPANTS, partBuf, MAXID*2) <= 0) {
	    xClose(lls);
            return ERR_XOBJ;
        }
    	partInternalize(fullPart,partBuf);
    	idLen = partToId(fullPart,idBuf,MAXID); 
	if (idLen < 1) {

	    xClose(lls);
            return ERR_XOBJ;
        }

        tagSessn = joinCreateTagSessn(self, hlpRcv, hlpType, lls, idBuf, idLen);
        if (tagSessn == ERR_XOBJ) {

            return ERR_XOBJ;
        }

        return tagSessn;
    }

    joinSessn = joinCreateSessn(self, hlpRcv, hlpType, tagSessn);
    if (joinSessn == ERR_XOBJ) {

        return ERR_XOBJ;
    }


    return joinSessn;
}

static XObj
joinCreateTagSessn (self, hlpRcv, hlpType, lls, idBuf, idLen)
    XObj        self, hlpRcv, hlpType, lls;
    char	*idBuf;
    int		idLen;
{
    XObj        s;
    TState      *ts;
    PState      *ps = (PState *)self->state;


    if (ps->tagIdLen == 0) {

    	ps->tagMap = mapCreate(JOIN_TAG_MAP_SZ, idLen);
        ps->tagIdLen = idLen;
    }

    if (idLen != ps->tagIdLen) {

        return ERR_XOBJ;
    }
       
    ts = X_NEW(TState);
    bzero((char *)ts, sizeof(TState));
    semInit(&ts->wait, 1);
    bcopy(idBuf,(char *)&ts->id,idLen);
    ts->numSeg = ps->numSeg;
    bcopy((char *) &ps->order, (char *)&ts->order, sizeof(ts->order));
 

    s = xCreateSessn(joinTagSessnInit, hlpRcv, hlpType, self, 1, &lls);
    s->rcnt=1;
    s->binding = (Bind) mapBind(ps->actMap, (char *)&lls, (int)s);
    if ( s->binding == ERR_BIND ) {

        xDestroy(s);
        xFree((char *) ts);
        return ERR_XOBJ;
      }
    ts->tagBinding = (Bind) mapBind(ps->tagMap, (char *)ts->id, (int)s);
    if ( ts->tagBinding == ERR_BIND ) {

	mapRemoveBinding(ps->tagMap,ts->tagBinding);
	s->binding = 0;
        xDestroy(s);
        xFree((char *) ts);
        return ERR_XOBJ;
      }

    s->state = (VOID *) ts;


    return(s);
}



static XObj
joinCreateSessn (self, hlpRcv, hlpType, tagSessn)
    XObj	self, hlpRcv, hlpType, tagSessn;
{
    XObj	s;
    SState	*ss;
    TState	*ts = (TState *)tagSessn->state;



    if (ts->state != JOININIT) {

	return ERR_XOBJ;
    }
    s = xCreateSessn(joinSessnInit, hlpRcv, hlpType, self, 1, &tagSessn);
    ss = X_NEW(SState);
    s->state = (VOID *) ss;
    ts->up[ts->next] = s;
    ss->seg = ts->next;
    ts->next++;
    tagSessn->rcnt++;
    if (ts->next == ts->numSeg) ts->state = JOINREADY;


    return s;
}


static xkern_return_t
joinOpenEnable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    int		idLen;
    char 	buf[MAXID];
    JEnable	*e;
    xkern_return_t val;



    idLen = partToId(part,buf, MAXID);

    if (ps->pasIdLen == 0) {

    	ps->pasMap = mapCreate(JOIN_PASSIVE_MAP_SZ, idLen);
        ps->pasIdLen = idLen;
    }

    if (idLen != ps->pasIdLen) {

        return XK_FAILURE;
    }

    if ( mapResolve(ps->pasMap, buf, &e) != XK_FAILURE ) {

        if ( e->next < ps->numSeg + 1) {

	    e->hlpRcv[e->next] = hlpRcv;
	    e->hlpType[e->next] = hlpType;
            e->next++;
	    e->rcnt++;
            return XK_SUCCESS;
        }

        return XK_FAILURE;
    }


    val = xOpenEnable(self, hlpType, xGetDown(self,0), part);
    if ( val == XK_FAILURE ) {

        return XK_FAILURE;
    }


    e = (JEnable *)xMalloc(sizeof(JEnable));
    e->saveHlpType = hlpType;
    e->hlpRcv[0] = hlpRcv;
    e->hlpType[0] = hlpType;
    e->next = 1;
    e->rcnt = 1;
    e->binding = mapBind(ps->pasMap, (char *)buf, (int) e);

    if (e->binding == ERR_BIND) {

      xFree((char *)e);
      return XK_FAILURE;
    }

    return XK_SUCCESS;

}


static xkern_return_t
joinOpenDisable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    int		idLen;
    char 	buf[MAXID];
    JEnable	*e;
    int 	i;
    XObj	hlp;

    



    idLen = partToId(part,buf, MAXID);

    if (idLen != ps->pasIdLen) {

        return XK_FAILURE;
    }

    if ( mapResolve(ps->pasMap, buf, &e) != XK_FAILURE ) {

        for (i=0; i<ps->numSeg + 1; i++) {
	    if ((e->hlpRcv[i] == hlpRcv) && (e->hlpType[i] == hlpType)) {

        	e->rcnt--;
	    	e->hlpRcv[i] = ERR_XOBJ;
	    	e->hlpType[i] = ERR_XOBJ;
	    }
  	}
       	if ( e->rcnt < 1 ) {
   	    mapRemoveBinding(ps->pasMap, e->binding);
	    hlp = e->saveHlpType;
	    xFree((char *)e);
    	    return xOpenDisable(self, hlp, xGetDown(self,0), part);
	} else {
    	    return XK_SUCCESS;
	}
    }
    return XK_FAILURE;

}

static xkern_return_t
joinOpenDone(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    XObj        tagSessn;
    XObj        joinSessn;
    PState      *ps;
    JEnable      *e;
    char	pBuf[MAXID*2];
    Part	p[5];
    char	idBuf[MAXID];
    int		idLen;
    xkern_return_t val;
    int		i;
   


    if ( self == hlpType ) {
        return XK_FAILURE;
    }

    ps = (PState *)self->state;

    if (xControl(lls,GETPARTICIPANTS,pBuf,MAXID*2) < 1) {
       return XK_FAILURE;
    }

    partInternalize(p,pBuf);
    idLen= part2ToId(p,idBuf,MAXID);
    
    if ((idLen != ps->pasIdLen) ||
        (mapResolve(ps->pasMap, &idBuf, &e) == XK_FAILURE )) {
        return XK_FAILURE;
    }

    xDuplicate(lls);

    idLen = partToId(p,idBuf,MAXID);

    tagSessn = joinCreateTagSessn(self,e->hlpRcv[0],e->hlpType[0], lls, idBuf, idLen);
    if (tagSessn == ERR_XOBJ) {
        xClose(lls);
        return XK_FAILURE;
    }
    val = xOpenDone(e->hlpRcv[0], tagSessn, self);
    if (val != XK_SUCCESS) {
        xClose(lls);
        return XK_FAILURE;
    }

    for (i=1; i<ps->numSeg +1; i++) {
        joinSessn = joinCreateSessn(self, e->hlpRcv[i], e->hlpType[i], tagSessn);
	if (joinSessn == ERR_XOBJ) {
            xClose(lls);
            return XK_FAILURE;
        }

        val = xOpenDone(e->hlpRcv[i], joinSessn, self);
	if (val != XK_SUCCESS) {
            xClose(lls);
            return XK_FAILURE;
        }
    }

    return val;
}



static xkern_return_t
joinDemux( self, lls, msg )
    XObj	self, lls;
    Msg		*msg;
{
  PState	*ps = (PState *)self->state;
  XObj 	s = 0;
  xkern_return_t val;
  


  if ( mapResolve(ps->actMap, &lls, &s) == XK_FAILURE ) { 

  	return XK_FAILURE;
  }

  val = xPop(s, lls, msg, 0);



  return val;
}


static xkern_return_t
joinClose( s )
    XObj	s;
{
    SState *ss;
    TState *ts;
    XObj tagSessn;



    s->rcnt = s->rcnt -1;
    if (s->rcnt > 0) {

	return XK_SUCCESS;
    }

    ss = (SState *)s->state;
    tagSessn = xGetDown(s,0);
    ts = (TState *)tagSessn->state;

    ts->state = JOINDYING;
    ts->up[ss->seg] = ERR_XOBJ;


    xClose(tagSessn);
    xDestroy(s);
    

    return XK_SUCCESS;
}


static xkern_return_t
joinTagClose( s )
    XObj	s;
{
    TState *ts = (TState *)s->state;
    PState *ps = (PState *)s->myprotl->state;


   
    ts->state = JOINDYING;
    s->rcnt = s->rcnt -1;
    if (s->rcnt > 0) {
	return XK_SUCCESS;
    }
    xAssert(ts->state == JOINDYING);

    mapRemoveBinding(ps->actMap,s->binding);
    mapRemoveBinding(ps->tagMap,ts->tagBinding);

    xClose(xGetDown(s, 0));
    xDestroy(s);

    return XK_SUCCESS;
}

static xmsg_handle_t
joinTagPush( s, msg )
    XObj	s;
    Msg 	*msg;
{
        return XK_FAILURE;
}

static xmsg_handle_t
joinPush( s, msg )
    XObj	s;
    Msg 	*msg;
{
    SState *ss;
    TState *ts;
    XObj   tagSessn;


    ss = (SState *)s->state;
    tagSessn = xGetDown(s,0);
    ts = (TState *)tagSessn->state;

    if (ts->state != JOINREADY) {

        return XK_FAILURE;
    }
    if (ts->downSegFull[ss->seg]) {
        return XK_FAILURE;
    }

    msgConstructCopy(&(ts->downSeg[ss->seg]),msg);
    ts->downSegFull[ss->seg] = 1;

    return XK_SUCCESS;
}

static xkern_return_t
joinPop( s, lls, m, h)
    XObj	s, lls;
    Msg 	*m;
    void        *h;
{
    return xDemux(s, m);
}

static xkern_return_t
joinTagPop( s, lls, m, h)
    XObj	s, lls;
    Msg 	*m;
    void        *h;
{
    TState	*ts = (TState *) s->state;
    xkern_return_t val;
    
    krb5_data        packet, packet2;
    krb5_kdc_req    *req;
    int              msg_len, retval;
    krb5_ap_req     *ap_req;
    krb5_data       *scratch;


    semWait(&ts->wait);



    /* the message is made up of a length and a packet */
    msg_len = msgLen(m);
#if 0
    /* testing for compatibility */
    if (msg_len < sizeof(int)) {

      return XK_FAILURE;
    }
    msgPop(m, mycopypop, &packet.length, sizeof(int), 0);


    if (msg_len != (packet.length + sizeof(int))) {

      return XK_FAILURE;
    }
#else
    packet.length = msg_len;
#endif 

    if ( !(packet.data = (char *)xMalloc(packet.length)) ) {
      return XK_FAILURE;
    }

    msgPop(m, mycopypop, (char*)packet.data, packet.length, 0);
    


    if (krb5_is_tgs_req(&packet)) {
      retval = decode_krb5_tgs_req(&packet, &req);
      if (retval) {
	return XK_FAILURE;
	/* free up memory */
      }

      msgConstructEmpty(&(ts->upSeg[0]));
      msgConstructBuffer(&(ts->upSeg[0]), (char*)req, sizeof(krb5_kdc_req));


      if (req->padata && req->padata[0] &&
	  (req->padata[0]->pa_type == KRB5_PADATA_AP_REQ)) {
	packet2.length = req->padata[0]->length;
	packet2.data = req->padata[0]->contents;

	retval = decode_krb5_ap_req(&packet2, &ap_req);
	if (retval) {
	  return XK_FAILURE;
	}



	/* re-encode the body of the request in order to perform */
	/* the checksum validation */
	retval = encode_krb5_kdc_req_body(req, &scratch);
	if (retval) {

	  return XK_FAILURE;
	}


	/* construct a message that is the encrypted part 
	   of the ticket */
	msgConstructEmpty(&(ts->upSeg[1]));
	msgConstructBuffer(&(ts->upSeg[1]),
			   ap_req->ticket->enc_part.ciphertext.data,
			   ap_req->ticket->enc_part.ciphertext.length);

	msgConstructEmpty(&(ts->upSeg[2]));
	msgConstructBuffer(&(ts->upSeg[2]),
			  ap_req->authenticator.ciphertext.data,
			  ap_req->authenticator.ciphertext.length);

	msgConstructEmpty(&(ts->upSeg[3]));
	msgConstructBuffer(&(ts->upSeg[3]), scratch->data, scratch->length);
	

	val = xPop(ts->up[ts->order[0]], lls, &ts->upSeg[0], 0);
	val = xPop(ts->up[ts->order[1]], lls, &ts->upSeg[1], 0);
	val = xPop(ts->up[ts->order[2]], lls, &ts->upSeg[2], 0);
	val = xPop(ts->up[ts->order[3]], lls, &ts->upSeg[3], 0);
	msgDestroy(&(ts->upSeg[0]));
	msgDestroy(&(ts->upSeg[1]));
	msgDestroy(&(ts->upSeg[2]));
	msgDestroy(&(ts->upSeg[3]));
	xFree((void*)ap_req);
	xFree((void*)req);
	xFree((void*)scratch);
      }
    } else if (krb5_is_as_req(&packet)) {

      /* I guess I should basically decode the message and pop */
      /* it up to the auth protocol. */
      retval = decode_krb5_as_req(&packet, &req);
      if (retval) {

	return XK_FAILURE;
	/* free up memory */
      }

#if 0
      /* REMOVE the next line */
      return XK_FAILURE;
#endif
      msgConstructEmpty(&(ts->upSeg[0]));
      msgConstructBuffer(&(ts->upSeg[0]), (char*)req, sizeof(krb5_kdc_req));
      val = xPop(ts->up[ts->order[0]], lls, &ts->upSeg[0], 0);
      msgDestroy(&(ts->upSeg[0]));
    } else {

#if 1
      /* REMOVE the next line */
      return XK_FAILURE;
#endif
    }
    

#if 0
    if (msgPop(m, joinHdrLoad1, (char *)&hdr, 8, 0) == FALSE) {

        return XK_SUCCESS;
    }

    if (msgPop(m, joinHdrLoad2, (char *)&hdr, 4*hdr.numSeg, 0) == FALSE) {
        return XK_SUCCESS;
    }

    for (i=0;i<hdr.numSeg-1; i++) {

	if (hdr.len[i]) { 
            msgConstructEmpty(&(ts->upSeg[i]));
            msgChopOff(m, &(ts->upSeg[i]),hdr.len[i]);
            ts->upSegFull[i] = 1;
	}
    }
    if (hdr.len[hdr.numSeg-1]) {  
        msgConstructCopy(&ts->upSeg[hdr.numSeg-1],m); 
        ts->upSegFull[hdr.numSeg-1] = 1;
    }

    for (i=0; i<hdr.numSeg; i++) {
        if (ts->upSegFull[ts->order[i]]) {
            val = xPop(ts->up[ts->order[i]],lls,&ts->upSeg[ts->order[i]],0);
            msgDestroy(&(ts->upSeg[ts->order[i]])); 
        }
      }

#endif 

   
    xControl(ts->controlSessn, JOINDONE, 0, 0);
    semSignal(&ts->wait);
    return val;
} 

static int
joinControlSessn(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsSession(s));

    return xControl(xGetDown(s,0), op, buf, len);
}

static int
joinTagControlSessn(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{
    TState *ts = (TState *) s->state;
    xAssert(xIsSession(s));


    switch (op) {

      case GETMAXPACKET:
      case GETOPTPACKET:
	{ 
	    int max;
	    if (len < sizeof(max)) return 0; 
	    if (xControl(xGetDown(s, 0), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
	    
	    *(int *)buf = (max - sizeof(JOINhdr));
	    return(sizeof(max));
	}

      case JOINGETNUMSEG:
        {   
	    if (len != sizeof(ts->numSeg)) return 0; 
	    *(int *)buf = ts->numSeg;
            return(sizeof(ts->numSeg));
        }
      case JOINGETORDER:
        {   
	    if (len != sizeof(ts->order)) return 0; 
	    bcopy((char *) &ts->order, buf, sizeof(ts->order));
            return(sizeof(ts->order));
        }
      case JOINSETCONTROL:
        {   
	    if (len != sizeof(ts->controlSessn)) return 0; 
	    bcopy(buf, (char *) &ts->controlSessn, sizeof(ts->controlSessn));
            return(sizeof(ts->controlSessn));
        }
      case JOINDONE:
        {
	    int retval;


	    retval = joindone_proc(s, op, buf, len);


            return retval;
        }

      default:
	return xControl(xGetDown(s,0), op, buf, len);
    }
}


static int
joinControlProtl( self, op, buf, len )
    XObj	self;
    int 	op, len;
    char 	*buf;
{
    PState *ps = (PState *) self->state;
    xAssert(xIsProtocol(self));
  
    switch (op) {
      case GETMAXPACKET:
      case GETOPTPACKET:
        {
            int max;
	    if (len < sizeof(max)) return 0; 
	    if (xControl(xGetDown(self, 0), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
	    *(int *)buf = (max - sizeof(JOINhdr));
            return(sizeof(max));
	}
      case JOINGETNUMSEG:
        {   
	    if (len != sizeof(ps->numSeg)) return 0; 
	    *(int *)buf = ps->numSeg;
            return(sizeof(ps->numSeg));
        }
      case JOINGETORDER:
        {   
	    if (len != sizeof(ps->order)) return 0; 
	    bcopy((char *) &ps->order, buf, sizeof(ps->order));
            return(sizeof(ps->order));
        }

      default: return xControl(xGetDown(self, 0), op, buf, len);
  }
  
}


static void
joinSessnInit(s)
    XObj s;
{

  s->push = joinPush;
  s->pop = joinPop;
  s->close = joinClose;
  s->control = joinControlSessn;
}

static void
joinTagSessnInit(s)
    XObj s;
{

  s->push = joinTagPush;
  s->pop = joinTagPop;
  s->close = joinTagClose;
  s->control = joinTagControlSessn;
}

static long mycopypop(dest, src, count, arg)
     void *dest;
     char *src;
     long count;
     void *arg;
{
  bcopy(src, (char*)dest, count);
  return count;
}


static int
partToId(p, buf, maxLen)
    Part *p;
    char *buf;
    int maxLen;
{
    int i,j,len;
    char *ptr;
    int numPart = partLen(p);
    u_char *ptr2;
    

    len = 0;
    ptr = (char *)buf;
    for (j=0;j<numPart; j++) {
    	for (i=p[j].stack.top; i>0; i--)  {
       		len += p[j].stack.arr[i-1].len;
		if (len > maxLen) return 0;
        	bcopy((char *)p[j].stack.arr[i-1].ptr,(char *)ptr, p[j].stack.arr[i-1].len);
        	ptr += p[j].stack.arr[i-1].len;
    	}
    }


    xIfTrace(JOINP, TR_EVENTS) {
        printf("part2ToLen - id= (");
        ptr2 = (u_char *)buf;
        for (i=0;i<len;i++) printf("%x",ptr2[i]);
        printf(")\n");
    }

    return(len);
}

static int
part2ToId(p, buf, maxLen)
    Part *p;
    char *buf;
    int maxLen;
{
    int i,len;
    char *ptr;
    u_char *ptr2;
    

    len = 0;
    ptr = (char *)buf;
    for (i=p[0].stack.top; i>0; i--)  {
       	len += p[1].stack.arr[i-1].len;
	if (len > maxLen) return 0;
       	bcopy((char *)p[1].stack.arr[i-1].ptr,(char *)ptr, p[1].stack.arr[i-1].len);
       	ptr += p[1].stack.arr[i-1].len;
    }

    xIfTrace(JOINP, TR_EVENTS) {
        printf("part2ToLen - id= (");
        ptr2 = (u_char *)buf;
        for (i=0;i<len;i++) printf("%x",ptr2[i]);
        printf(")\n");
    }

    return(len);
}

static int
isOrderValid(order,numseg)
int *order;
int numseg;
{
    int i,j;
    int found;
    
    for (i=0; i<numseg;i++) {
       found=0;
       for (j=0;j<numseg;j++) {
           if (order[j] == i) found = 1; 
       }
       if (!found) return 0;
    }
    return 1;
}


/*
 * Sets the 'port' field of the protocol state
 */
static void
processRomFile( self )
    XObj        self;
{
    PState      *ps = (PState *)self->state;
    int         i,k;

    for ( i=0; rom[i][0]; i++ ) {
        if ( ! strncmp(rom[i][0], self->name, strlen(self->name)) ) {
            if ( ! strcmp(rom[i][0], self->fullName) ) {
                if ( rom[i][1] ) {
                    if ( ! strcmp(rom[i][1], "segments") && 
			   rom[i][2] &&
                           (sscanf(rom[i][2], "%d", &ps->numSeg) >= 1)
		        ) {
                        continue;
                    }
                    if ( ! strcmp(rom[i][1], "order")) {
 			for (k=2;(k<ps->numSeg+2) && rom[i][k];k++) {
                            if (sscanf(rom[i][k], "%d", &ps->order[k-2]) <1 ) {
            			sprintf(errBuf, "%s format error in line %d of the rom file",
                    			self->fullName, i + 1);
            			xError(errBuf);
            			break;
			    }
			}
                        continue;
                    }
                }
            } else {
                continue;
            }
            sprintf(errBuf, "%s format error in line %d of the rom file",
                    self->fullName, i + 1);
            xError(errBuf);
            break;
        }
    }
}
static int
joindone_proc(s, op, buf, bufsize)
    XObj	s;
    int 	op, bufsize;
    char 	*buf;
{
    TState         *ts = (TState *) s->state;
    Msg             fullmsg;
    krb5_kdc_rep    rep;
    krb5_data      *scratch;
    int             i, retval;

#define reset() for (i=0;i<ts->numSeg; i++) {      \
                  if (ts->downSegFull[i])          \
                    msgDestroy(&(ts->downSeg[i])); \
                    ts->downSegFull[i] = 0;        \
                 }

    if (ts->downSegFull[0] && ts->downSegFull[1] && ts->downSegFull[2]) {
      if ( msgLen(&(ts->downSeg[0])) != sizeof(krb5_kdc_rep) ) {
	reset();
	return -1;
      }
      msgPop( &(ts->downSeg[0]), mycopypop, (char*)&rep,
	      sizeof(krb5_kdc_rep), 0 );
      rep.enc_part.ciphertext.length = msgLen(&(ts->downSeg[2]));
      rep.enc_part.ciphertext.data = xMalloc(rep.enc_part.ciphertext.length);
      if (!rep.enc_part.ciphertext.data) {
	reset();
	return -1;
      }
      msgPop(&(ts->downSeg[2]), mycopypop, (char*)rep.enc_part.ciphertext.data,
	     rep.enc_part.ciphertext.length, 0);

      rep.ticket->enc_part.ciphertext.length = msgLen(&(ts->downSeg[1]));
      rep.ticket->enc_part.ciphertext.data = xMalloc(rep.ticket->enc_part.ciphertext.length);
      if (!rep.ticket->enc_part.ciphertext.data) {

	reset();
	xFree(rep.enc_part.ciphertext.data);
	return -1;
      }
      msgPop(&(ts->downSeg[1]), mycopypop,
	     (char*)rep.ticket->enc_part.ciphertext.data,
	     rep.ticket->enc_part.ciphertext.length, 0);

      if (rep.msg_type == KRB5_AS_REP) {

	/* following line added just to see if it encodes ok */
	/* rep.enc_part.etype = ETYPE_DES_CBC_MD5; */
	rep.enc_part.etype = 1;
	retval = encode_krb5_as_rep(&rep, &scratch);
      } else if (rep.msg_type == KRB5_TGS_REP) {

	retval = encode_krb5_tgs_rep(&rep, &scratch);
      } else {
	retval = -1;
      }

      if (retval) {
	reset();
	xFree(rep.ticket->enc_part.ciphertext.data);
	xFree(rep.enc_part.ciphertext.data);
	return -1;
      }

      msgConstructBuffer(&fullmsg, (char *)scratch->data, scratch->length);
      /* msgPush(&fullmsg, mycopypush, (char *)&scratch->length,sizeof(int),0); */

      xPush(xGetDown(s, 0), &fullmsg);
    
      msgDestroy(&fullmsg);
      reset();
      xFree(rep.ticket->enc_part.ciphertext.data);
      xFree(rep.enc_part.ciphertext.data);
      return 0;
    }
    reset();
#undef reset
    
    return -1;
}


    
    
    
static void mycopypush(src, dst, count, arg)
     void *src;
     char *dst;
     long  count;
     void *arg;
{
  bcopy((char *)src, dst, count);
}



