/*     
 * $RCSfile: composer2.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: composer2.c,v $
 * Revision 1.1  1994/04/18 22:02:03  slm
 * Initial revision
 *
 */

#include "xkernel.h"
#include "psync.h"
#include "divider.h"
#include "composer2_internal.h"

       void           composer_init(XObj);
static void           getproc_protl(XObj);
static XObj           composer_open(XObj, XObj, XObj, Part *);
static xkern_return_t composer_openenable(XObj, XObj, XObj, Part *);
static xkern_return_t composer_opendone(XObj, XObj, XObj, XObj);
static int            composer_controlprotl(XObj, int, char *, int);

static int my_composer_id = 0;
       int tracecomposerp = 1;

void composer_init(self)
XObj self;
{
    COMPOSERstate *pstate;

    xAssert(xIsProtocol(self));
    xTrace0(composerp, TR_GROSS_EVENTS, "COMPOSER init");
    if (!xIsProtocol(xGetDown(self,0)))
        Kabort("COMPOSER: could not get lower protocol DIVIDER");
    if (!xIsProtocol(xGetDown(self,1)))
	Kabort("COMPOSER: could not get lower protocol DISPATCH");

    getproc_protl(self);

    if ((pstate = (COMPOSERstate *)xMalloc(sizeof(COMPOSERstate))) == NULL)
	Kabort("COMPOSER: malloc failed in composer_init");
    self->state = (char *)pstate;
    pstate->map            = mapCreate(100, sizeof(int));
    pstate->map_openenable = mapCreate(100, sizeof(int));
}

static void getproc_protl(self)
XObj self;
{
    xAssert(xIsProtocol(self));

    self->open       = composer_open;
    self->openenable = composer_openenable;
    self->opendone   = composer_opendone;
    self->control    = composer_controlprotl;
}

static XObj composer_open(self, hlpRcv, hlpType, p)
XObj self, hlpRcv, hlpType;
Part *p;	/* p = [opcode,port,psync1,...,psyncN,NULL] */
{
    COMPOSERstate *pstate = (COMPOSERstate *)self->state;
    int           opcode, port, i;
    XObj          psync_sessn, dispatch_sessn;
    Part          part[MAX_PART+3];
    PSYNCaddr     *curr_addr;
    F_msg         fmsg;
    Msg           msg;

    xAssert(xIsProtocol(self));
    xTrace0(composerp, TR_FUNCTIONAL_TRACE, "COMPOSER: entering composer_open");

    /* Get the opcode and the port number from the participant list. */
    opcode = *(int *)partPop(p[0]);
    port   = *(int *)partPop(p[1]);
    xTrace2(composerp, TR_DETAILED, "COMPOSER: opcode = %d, port = %d",
	    opcode, port);

    /*
     * If a DIVIDER's xOpen has been called by this composer, an entry should
     * exist for the given port number.
     */
    /* `map' is <port> --> <psync_sessn> */
    if (mapResolve(pstate->map, (VOID *)&port, (VOID **)&psync_sessn) ==
	    XK_FAILURE) {
        /*
         * There is no DIVIDER session opened for this port number; one must
	 * be opened.
         */
        partInit(part, MAX_PART+3);
        partPush(part[0], (VOID *)&my_composer_id, sizeof(int));
        partPush(part[1], (VOID *)&port, sizeof(int));

        for (i = 2; curr_addr = (PSYNCaddr *)partPop(p[i]); i++) {
            xTrace3(composerp, TR_DETAILED,
                    "COMPOSER: part %d, PSYNCport = %d, IPhost = %s",
                    i-2, curr_addr->port, ipHostStr(&curr_addr->host));
            partPush(part[i], (VOID *)curr_addr, sizeof(PSYNCaddr));
        }
	partPush(part[i], (VOID *)NULL, 0);

        /* DIVIDER returns PSYNC session instead of its own session. */
	/* *** UNCONVENTIONAL *** */
        psync_sessn = xOpen(self, self, xGetDown(self,0), part);

        /*
         * Bind the PSYNC session returned by DIVIDER to the current port
	 * number, a mapping of PSYNC sessionn to port_id.
         */
        mapBind(pstate->map, (VOID *)&port, (VOID *)psync_sessn);
    }
    /*
     * Put the PSYNC session (returned from opening DIVIDER) and the opcode in
     * the part list for DISPATCH.
     */
    partInit(part, 2);
    partPush(part[0], (VOID *)&psync_sessn, sizeof(XObj));
    partPush(part[1], (VOID *)&opcode, sizeof(int));

    /* Open the DISPATCH protocol. */
    dispatch_sessn = xOpen(self, self, xGetDown(self,1), part);

    /*
     * Set DISPATCH's up pointer to directly point to the user protocol
     * above me.  *** UNCONVENTIONAL ***
     */
    xSetUp(dispatch_sessn, hlpRcv);

    /*
     * Once for ever user process, send a STARTUP msg to the other replicas so
     * that they can set up the connections.
     *
     * A user process (server) may support several opcodes.  The user process
     * should use opcode = 1 only once to trigger this code.
     */
    if (opcode == 1) {
        /* Send the first msg which will set up all the connections. */
        if (xControl(psync_sessn, PSYNC_LOOPBACK, (char *)NULL, 0) == -1)
	    Kabort("TOTAL_ORDER: PSYNC_LOOPBACK returned error");
        bzero((char *)&fmsg, sizeof(F_msg));
#ifdef XK_FIRST_INT
	fmsg.first       = 1;
#else
        fmsg.first       = '1';
#endif
        fmsg.port        = port;
        fmsg.composer_id = my_composer_id;
        msgConstructBuffer(&msg, (char *)&fmsg, sizeof(F_msg));
        xTrace3(composerp, TR_DETAILED,
       "COMPOSER: pushing first msg, port = %d, composer_id = %d, msgLen = %ld",
                port, my_composer_id, msgLen(&msg));
        xPush(psync_sessn, &msg);
	msgDestroy(&msg);
    }
    xTrace0(composerp, TR_FUNCTIONAL_TRACE, "COMPOSER: exiting composer_open");
    return dispatch_sessn;
}

static xkern_return_t composer_openenable(self, hlpRcv, hlpType, p)
XObj self, hlpRcv, hlpType;
Part *p;	/* p = [opcode,port,psync1] */
{
    COMPOSERstate *pstate = (COMPOSERstate *)self->state;
    int           opcode, port;
    PSYNCaddr     *psyncaddr;
    OPENvalue     *openvalue, *openval;
    Part          part[2];

    xAssert(xIsProtocol(self));
    xAssert(xIsProtocol(hlpRcv));
    xTrace0(composerp, TR_FUNCTIONAL_TRACE,
	    "COMPOSER: entering composer_openenable");

    /* `openval' is a linked list of <psyncaddr,opcode,hlp>. */

    opcode    = *(int *)partPop(p[0]);
    port      = *(int *)partPop(p[1]);
    psyncaddr = (PSYNCaddr *)partPop(p[2]);
    xTrace4(composerp, TR_DETAILED,
	    "COMPOSER: opcode = %d, port = %d, PSYNCport = %d, IPhost = %d",
	    opcode, port, psyncaddr->port, ipHostStr(&psyncaddr->host));

    if ((openvalue = (OPENvalue *)xMalloc(sizeof(OPENvalue))) == NULL) {
        xError("COMPOSER: malloc failed in composer_openenable");
	return XK_FAILURE;
    }
    else {
        openvalue->psyncaddr = *psyncaddr;
        openvalue->opcode    = opcode;
        openvalue->hlp       = hlpRcv;
    }

    /* `map_openenable' is <port> --> <psyncaddr,opcode,hlp> */
    if (mapResolve(pstate->map_openenable, (VOID *)&port,
		   (VOID **)&openval) == XK_FAILURE) {
        /*
         * No openenable done for this port.  Bind the current port number to
	 * the psyncaddr, opcode and hlp combination to be used with the
	 * opendone combination.
         */
        openvalue->next = NULL;
        mapBind(pstate->map_openenable, (VOID *)&port, (VOID *)openvalue);
        /*
         * Prepare part list as composer_id and psyncaddr and call DIVIDER's
	 * openenable.
         */
        partInit(part, 2);
        partPush(part[0], (VOID *)&my_composer_id, sizeof(int));
        partPush(part[1], (VOID *)psyncaddr, sizeof(PSYNCaddr));

        /*
         * Do an openenable on DIVIDER.  (DISPATCH and TOTAL_ORDER are always
	 * opened actively.)
         */
        if (xOpenEnable(self, self, xGetDown(self,0), part) == XK_FAILURE) {
	   xError("COMPOSER: error in openenable");
	   return XK_FAILURE;
	}
    }
    else {
        /* Openenable already done for this port; add new openvalue to list. */
        openvalue->next = openval->next;
        openval->next   = openvalue;
    }
    xTrace0(composerp, TR_FUNCTIONAL_TRACE,
	    "COMPOSER: exiting composer_openenable");
    return XK_SUCCESS;
}

static xkern_return_t composer_opendone(self, sessn, llp, hlpType)
XObj self, sessn, llp, hlpType;
/* sessn is actually the PSYNC session handed over to me by DIVIDER */
{
    COMPOSERstate *pstate = (COMPOSERstate *)self->state;
    int           port;
    Part          part[2];
    OPENvalue     *openval;
    XObj          dispatch_sessn;

    xAssert(xIsProtocol(self));
    xTrace0(composerp, TR_FUNCTIONAL_TRACE,
	    "COMPOSER: entering composer_opendone");

    /*
     * PSYNC and DIVIDER have set up connections.  Get the port number for the
     * current msg.
     */
    if (xControl(xGetDown(self,0), DIVIDER_GETCURRPORT, (char *)&port,
	         sizeof(port)) == -1) {
	xError("COMPOSER: DIVIDER_GETCURRPORT returned error");
	return XK_FAILURE;
    }

    /*
     * Bind the sessn (which is a PSYNC sessn) to the port.  This is a special
     * type of GETPART in which a part list for the port is prepared and passed
     * up.
     */
    xTrace2(composerp, TR_DETAILED,
            "COMPOSER: binding port %d to PSYNC session %d", port, sessn);
    mapBind(pstate->map, (VOID *)&port, (VOID *)sessn);

    /*
     * For the port number, find the psyncaddr, opcode and hlp that have done
     * an openenable.
     */
    /* `map_openenable' is <port> --> <psyncaddr,opcode,hlp> */
    if (mapResolve(pstate->map_openenable, (VOID *)&port,
		   (VOID **)&openval) == XK_FAILURE) {
        xError("COMPOSER: cannot resolve port to openval in opendone");
	return XK_FAILURE;
    }

    partInit(part, 2);
    partPush(part[0], (VOID *)&sessn, sizeof(XObj));
    while (openval) {
        /*
	 * For each entry in the OPENvalue list <psyncaddr,opcode,hlp>, open
	 * DISPATCH with [psync,opcode], (in other words, for each hlp that's
	 * done an openenable for this opcode, open the DISPATCH session
	 * on its behalf).  Then call opendone.
         */
        partPush(part[1], (VOID *)&openval->opcode, sizeof(int));
        dispatch_sessn = xOpen(self, self, xGetDown(self,1), part);

        /*
         * Set DISPATCH's up pointer to directly point to the user protocol
         * above me.  *** UNCONVENTIONAL ***
         */
        xSetUp(dispatch_sessn, openval->hlp);

        /*
         * Call opendone of the hlp (USER) with the [opcode] part list and
	 * hand over the DISPATCH session to it (this sets DISPATCH session's
	 * up to the USER protocol).
         */
        if (xOpenDone(openval->hlp, dispatch_sessn, self) == XK_FAILURE) {
	   xError("COMPOSER: error in opendone");
	   return XK_FAILURE;
	}
        openval = openval->next;
    }
    if (xControl(sessn, PSYNC_LOOPBACK, (char *)NULL, 0) == -1)
	Kabort("TOTAL_ORDER: PSYNC_LOOPBACK returned error");
    xTrace0(composerp, TR_FUNCTIONAL_TRACE,
	    "COMPOSER: exiting composer_opendone");
    return XK_SUCCESS;
}

static int composer_controlprotl(self, opcode, buf, len)
XObj self;
int  opcode, len;
char *buf;
{
    xAssert(xIsProtocol(self));

    switch (opcode) {
        default:
            xError("COMPOSER: in default case of controlprotl");
            return -1;
    }
}
