/*     
 * $RCSfile: rip.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: rip.c,v $
 * Revision 1.2  1994/09/16 16:27:21  menze
 * Fixed syntax error
 * 'SEC' renamed 'ONESEC' to avoid name conflicts
 *
 */

#include "xkernel.h"
#include "ip.h"
#include "vnet.h"
#include "arp.h"
#include "udp.h"
#include "riproute.h"
#include "riproute_i.h"
#include "rip.h"
#include "rip_i.h"

int	traceripp;

static IPhost	ip_anyaddr = {0xff,0xff,0xff,0xff};
static IPhost	ip_maskall = {0xff,0xff,0xff,0xff};
static IPhost   ipNull = { 0, 0, 0, 0 };



#ifdef __STDC__

static void		getproc_protl(XObj);
static xkern_return_t	ripCloseProtl( XObj );
static int		ripControlProtl( XObj, int, char *, int ) ;
static xkern_return_t	ripDemux( XObj, XObj, Msg * );
static long		ripHdrLoad( void *, char *, long, void * );
static void		ripHdrStore( void *, char *, long, void * );
int ripcheck(struct rip *, int);
int riprecv(XObj, struct rip *, int, IPhost, int);
int ripok(struct riprt *);
int riprepl(XObj, struct rip *, int, IPhost, unsigned short, XObj);
int ripsend(XObj, IPhost, unsigned short, XObj);
int ripsendrequest(XObj, XObj);
int ripifset(XObj, struct rq *,IPhost, unsigned short, XObj);
int ripadd(XObj, struct rq *, struct route *);
int netnum(IPhost *, IPhost *);
int ripstart(struct rq *);
void ripclock(Event, VOID *);
void ripout(XObj);
int ripmetric(struct route *, int);
static void             ripHdrDisplay(struct rip *, int);
void ripprintentry(route *);
void ripprint(XObj);
void ripage(XObj);




#else

static void		getproc_protl();
static xkern_return_t	ripCloseProtl();
static int		ripControlProtl() ;
static xkern_return_t	ripDemux();
static long		ripHdrLoad();
static void		ripHdrStore();
int ripcheck();
int riprecv();
int ripok();
int riprepl();
int ripsend();
int ripsendrequest();
int ripifset();
int ripadd();
int netnum();
int ripstart();
void ripout();
void ripclock();
int ripmetric();
static void             ripHdrDisplay();
void ripprintentry();
void ripprint();
void ripage();

#endif /* __STDC__ */


/*
 * ripHdrStore -- write header to potentially unaligned msg buffer.
 * Note:  *hdr will be modified
 */
static void
ripHdrStore(hdr, dst, len, arg)
    VOID *hdr;
    char *dst;
    long int len;
    VOID *arg;
{
    struct rip *packet = hdr;
    int i;

    xTrace0(ripp, TR_EVENTS, "ripHdrStore entered");

    for (i =0; i< (len-RIPHSIZE)/sizeof(struct riprt); i++) {
	packet->rip_rts[i].rr_family = htons(packet->rip_rts[i].rr_family);
	packet->rip_rts[i].rr_metric = htonl(packet->rip_rts[i].rr_metric);
    }
    bcopy((char *)packet,dst,len);

}


/*
 * ripHdrLoad -- load header from potentially unaligned msg buffer.
 * Result of checksum calculation will be in hdr->sum.
 */
static long
ripHdrLoad(hdr, src, len, arg)
    VOID *hdr;
    char *src;
    long int len;
    VOID *arg;
{
    struct rip *packet = hdr;
    int num_routes;
    int i;
   
    xTrace0(ripp, TR_EVENTS, "ripHdrLoad entered");

    num_routes = (len-RIPHSIZE)/sizeof(struct riprt);
 
    if (num_routes < 1) {
        xTrace0(ripp, TR_ERRORS, "ripHdrLoad no routes");
	return 0;
    }

    if ((len-RIPHSIZE) % sizeof(struct riprt) != 0) {
        xTrace0(ripp, TR_ERRORS, "ripHdrLoad not right size");
	return 0;
    }

    bcopy(src, (char *)packet,len);

    for (i =0; i< num_routes; i++) {
        packet->rip_rts[i].rr_family = ntohs(packet->rip_rts[i].rr_family);
        packet->rip_rts[i].rr_metric = ntohl(packet->rip_rts[i].rr_metric);
    }   

    return len;
}


/*
 * rip_init
 */
void
rip_init(self)
    XObj self;
{
    Part	part[2];
    PSTATE	*pstate;
    XObj	llp;
    long	rip_port = UP_RIP;
    IPhost	addrs[RIPMAXINTER];
    Interface	*ifc;
    IPhost	mask;
    int 	i;
    IProuteInfo ipinfo;
    
    xTrace0(ripp, TR_GROSS_EVENTS, "RIP init");
    xAssert(xIsProtocol(self));
    
    getproc_protl(self);
    pstate = X_NEW(PSTATE);
    bzero((char *)pstate, sizeof(PSTATE));
    self->state = (VOID *)pstate;

    pstate->active = 0;
    pstate->self = self;
    semInit(&pstate->riplock,1);

    ript_init(pstate,&ipNull);


    if ( ! xIsProtocol(llp = xGetDown(self, 0)) ) {
	Kabort("RIP could not get lower protocol");
    }

    if ( xControl(llp,GETMYHOSTCOUNT, (char *)&pstate->numNif,sizeof(int)) 
         != sizeof(int)) {
	Kabort("RIP could not get number of interfaces");
    }

    if (pstate->numNif > 1) {
        pstate->gateway = TRUE;
    }

    if ( xControl(llp,GETMYHOST,(char *)addrs,pstate->numNif*sizeof(IPhost))         != pstate->numNif*sizeof(IPhost)) {
	Kabort("RIP could not get interface ip address");
    }

    for (i=0;i<pstate->numNif;i++) {
	ifc = &pstate->nif[i];
	ifc->host = addrs[i];
        netMaskFind(&ifc->mask, &ifc->host);
        netMaskDefault(&mask, &ifc->host);
        ifc->subnet = ! IP_EQUAL(mask, ifc->mask);
	IP_AND(ifc->bcast, ifc->host, ifc->mask);
	IP_COMP(mask,ifc->mask);
	IP_AND(mask, mask, ip_anyaddr );
	IP_OR(ifc->bcast, mask, ifc->bcast);
	IP_AND(ifc->net, ifc->mask, ifc->host);
        partInit(part, 2);
        partPush(part[0],&ifc->bcast , sizeof(IPhost));
        partPush(part[0],&rip_port , sizeof(rip_port));
        partPush(part[1],&ifc->host, sizeof(IPhost));
        partPush(part[1],&rip_port , sizeof(rip_port));
    	ifc->bcast_s = xOpen(self, self, xGetDown(self, 0), part);
        if ( ifc->bcast_s == ERR_XOBJ ) {
	    xTrace1(ripp, TR_ALWAYS, "rip_init: can't open bcast session %d",i);
	    Kabort("");
	}
	ifc->active = 1;
        if ( xControl(ifc->bcast_s,VNET_GETINTERFACEID, (char *)&ifc->vnet_id, sizeof(VOID *)) != sizeof(VOID *)) {
            xTrace0(ripp, TR_EVENTS, "RIP init: could not get interface id");
	    Kabort("");
        }

	ript_add(pstate,&ifc->net,&ifc->mask,&ifc->host,0, RIPROUTE_TTL_INFINITE ,i);
	ipinfo.net = ifc->net;
	ipinfo.mask = ifc->mask;
	ipinfo.gw = ifc->host;
	ipinfo.metric = 0;
	ipinfo.ttl = RIPROUTE_TTL_INFINITE;
        if ( xControl(llp,IP_CHANGEROUTE,(char *)&ipinfo,sizeof(ipinfo)) 
             < 0 ) {
	      Kabort("RIP could not set route");
        }
    }


    

    /* openEnable UDP session */
    partInit(part, 1);
    partPush(part[0],ANY_HOST,0);
    partPush(part[0],&rip_port , sizeof(rip_port));
    if ( xOpenEnable(self, self, llp, part) == XK_FAILURE ) {
	xTrace0(ripp, TR_ALWAYS, "rip_init: can't openenable udp protocol");
	xFree((char *) pstate);
	return;
    }

    pstate->active = 1;


    pstate->ageInterval = RIPAGEINT;
    pstate->ageTime = RIPAGEINT;
    pstate->printInterval = RIPPRINTINT;
    pstate->printTime = RIPPRINTINT;
    pstate->outputInterval = RIPINT;
    pstate->outputTime = 0;
    pstate->outputForced = 1;

    evDetach( evSchedule( ripclock, self, 0));

    /* speed up process by requesting routing table of all routers */
    for (i=0;i<pstate->numNif;i++) {
	ifc = &pstate->nif[i];
        ripsendrequest(self,ifc->bcast_s);
    }

    xTrace0(ripp, TR_GROSS_EVENTS, "RIP init done");
}


/*
 * ripControlProtl
 */
static int
ripControlProtl(self, opcode, buf, len)
    XObj self;
    int opcode;
    char *buf;
    int len;
{

    xAssert(xIsProtocol(self));
    
             
    xTrace0(ripp, TR_EVENTS, "ripControlProtl entered");

    switch (opcode) {
	
      default:
	return xControl(xGetDown(self, 0), opcode, buf, len);
	
    }
}


/*
 * ripCloseProtl
 */
static xkern_return_t
ripCloseProtl(self)
    XObj self;
{
  PSTATE        *pstate;
  
  xAssert(xIsProtocol(self));
  
  xTrace0(ripp, TR_EVENTS, "ripCloseProtl entered");

  if (((self->rcnt)) <= 0) {
      pstate = (PSTATE *) self->state;
  }
  return XK_SUCCESS;
}


/*
 * ripDemux
 */
static xkern_return_t
ripDemux(self, lls, dg)
    XObj self;
    XObj lls;
    Msg *dg;
{
    struct	rip	prip;
    PSTATE      *pstate;
    int		len;
    Part	p[2];
    long	*gwPort = 0;
    long	*myPort =  0;
    IPhost	*gwHost = 0;
    IPhost	*myHost = 0;
    char	partBuf[256];
    int		numParts;
    VOID	*vnet_id;
    int		ifnum;
    int		i;
    
    pstate = (PSTATE *)self->state;
    xTrace0(ripp, TR_EVENTS, "RIP Demux");

    bzero((char *)&prip,sizeof(prip));
    len = msgLen(dg);
    msgPop(dg, ripHdrLoad, &prip, len, 0);


    xIfTrace(ripp, TR_EVENTS) {
        ripHdrDisplay(&prip, len);
    }


    if (ripcheck(&prip, len) != OK) {
         return XK_SUCCESS;
    } 

    if (xControl(lls,GETPARTICIPANTS, partBuf, sizeof(partBuf)) < 0 ) {
        xTrace0(ripp, TR_EVENTS, "RIP Demux: could not get participants");
        return XK_SUCCESS;
    }

    numParts = partExtLen(partBuf);
    if ( numParts != 2 ) {
        return XK_SUCCESS;
    }

    partInternalize(p, partBuf);

    gwPort = (long *)partPop(p[0]);
    gwHost = (IPhost *)partPop(p[0]);
    myPort = (long *)partPop(p[1]);
    myHost = (IPhost *)partPop(p[1]);

    if ( xControl(lls,VNET_GETINTERFACEID, (char *)&vnet_id, sizeof(VOID *))
         != sizeof(VOID *)) {
        xTrace0(ripp, TR_EVENTS, "RIP Demux: could not get interface id");
        return XK_SUCCESS;
    }

    ifnum = -1;
    for (i=0; i<pstate->numNif; i++) {
	if (pstate->nif[i].vnet_id == vnet_id) {
	    ifnum = i;
	}
    }

    xTrace4(ripp, TR_EVENTS,"ifnum  = %d infHost = %s gwHost = %s myHost = %s", ifnum, ipHostStr(&pstate->nif[ifnum].host), ipHostStr(gwHost), ipHostStr(myHost));
 

    if (IP_EQUAL(*gwHost, pstate->nif[ifnum].host)) {
        xTrace0(ripp, TR_EVENTS, "RIP Demux: loopback detected");
        return XK_SUCCESS;
    }
	 
    if ( ifnum == -1) { 
        xTrace0(ripp, TR_EVENTS, "RIP Demux: could not match interface id");
        return XK_SUCCESS;
    }
    
    
    switch (prip.rip_cmd) {
    case RIP_RESPONSE:
        if (*gwPort == UP_RIP) {
            riprecv(self, &prip, len, *gwHost, ifnum);
        }
        break;
    case RIP_REQUEST:
        if (pstate->gateway || *gwPort != UP_RIP) {
            riprepl(self, &prip, len, *gwHost, *gwPort, lls);
        }
        break;
    default:
        break;
     }

    return XK_SUCCESS;
}


static void
getproc_protl(s)
    XObj s;
{
  xAssert(xIsProtocol(s));
  s->close = ripCloseProtl;
  s->control = ripControlProtl;
  s->demux = ripDemux;
}



/*------------------------------------------------------------------------
 *  ripcheck  -  check a RIP packet for proper format
 *------------------------------------------------------------------------
 */
int ripcheck(prip, len)
struct	rip	*prip;
int		len;
{
	int	i, j, nrts;

        xTrace0(ripp, TR_EVENTS, "ripcheck entered");


	switch (prip->rip_vers) {
	case	0:			/* never OK		*/
		return SYSERR;
	case	1:			/* more checks below	*/
		break;
	default:			/* >1 always ok		*/
		return OK;
	}
	/* check all "must be zero" fields */

	if (prip->rip_mbz)
		return SYSERR;
	nrts = (len - RIPHSIZE)/sizeof(struct riprt);
	for (i=0; i<nrts; ++i) {
		struct	riprt	*prr = &prip->rip_rts[i];

		if (prr->rr_mbz)
			return SYSERR;
		for (j=1; j<RIPNULLADDR; ++j)
			if (prr->rr_mbz2[j])
				return SYSERR;
	}
	return OK;	/* this one's ok in my book... */
}


/*------------------------------------------------------------------------
 *  riprecv  -  process a received RIP advertisement
 *------------------------------------------------------------------------
 */
int riprecv(self,prip, len, gw, ifnum)
XObj		self;
struct	rip	*prip;
int		len;
IPhost		gw;
int		ifnum;
{
        PSTATE 	*pstate = (PSTATE *)self->state;
	IPhost		mask, addr;
	int		nrts, rn;
	IProuteInfo	ipinfo;

        xTrace0(ripp, TR_EVENTS, "riprecv  entered");

	nrts = (len - RIPHSIZE)/sizeof(struct riprt);
	semWait(&pstate->riplock); /* prevent updates until we're done	*/
	for (rn=0; rn<nrts; ++rn) {
		struct riprt	*rp = &prip->rip_rts[rn];
		if (!ripok(rp))
			continue;
		addr = rp->rr_addr;
		netMaskFind(&mask, &addr);
		ript_add(pstate, &addr, &mask, &gw, rp->rr_metric, RIPRTTL, ifnum);
        	ipinfo.net = addr;
        	ipinfo.mask = mask;
        	ipinfo.gw = gw;
        	ipinfo.metric = rp->rr_metric;
        	ipinfo.ttl = RIPRTTL;
        	if ( xControl(xGetDown(self,0),IP_CHANGEROUTE,(char *)&ipinfo,sizeof(ipinfo))
        	     < 0 ) {
        	      Kabort("RIP could not set route");
        	}
	}
	semSignal(&pstate->riplock);
	return OK;
}


/*------------------------------------------------------------------------
 *  ripok  -  determine if a received RIP route is ok to install
 *------------------------------------------------------------------------
 */
int ripok(rp)
struct	riprt	*rp;
{

        xTrace0(ripp, TR_EVENTS, "ripok  entered");

	if (rp->rr_family != AF_INET)
		return FALSE;
	if (rp->rr_metric > RIP_INFINITY)
		return FALSE;
	if (IP_CLASSD(rp->rr_addr) || IP_CLASSE(rp->rr_addr))
		return FALSE;
	if (rp->rr_addr.a == 0 &&
	    !bcmp((char *)&rp->rr_addr,(char *) &ip_anyaddr, IP_ALEN))
		return FALSE;		/* net 0, host non-0		*/
	if (rp->rr_addr.a == 127)
		return FALSE;		/* loopback net			*/
	return TRUE;
}



/*------------------------------------------------------------------------
 *  riprepl  -  process a received RIP request
 *------------------------------------------------------------------------
 */
int riprepl(self,pripin, len, gw, port, udp_s)
XObj		self;
struct	rip	*pripin;
int		len;
IPhost		gw;
unsigned short	port;
XObj		udp_s;
{
	PSTATE 	*pstate = (PSTATE *)self->state;
	struct	rip	prip;
	struct	route	prt;
	int		rn, nrts;
	Msg		msg;

        xTrace0(ripp, TR_EVENTS, "riprepl entered");

	nrts = (len - RIPHSIZE)/sizeof(struct riprt);
	if (nrts == 1 && pripin->rip_rts[0].rr_family == 0 &&
	    pripin->rip_rts[0].rr_metric == RIP_INFINITY) {
		return ripsend(self, gw, port, udp_s);	/* send the full table	*/
        }
	bcopy((char *)&pripin, (char *)&prip, len);
	for (rn = 0; rn < nrts; ++rn) {
		struct riprt	*rp = &prip.rip_rts[rn];

		if (rp->rr_family != AF_INET)
			continue;
		if (ript_get(&pstate->rtTbl, &rp->rr_addr, &prt) == XK_SUCCESS) {  
			rp->rr_metric = prt.metric;
		} else {
			rp->rr_metric = RIP_INFINITY;
		}
	}
	prip.rip_cmd = RIP_RESPONSE;
	prip.rip_vers = RIP_VERSION;
	prip.rip_mbz = 0;
        xIfTrace(ripp, TR_EVENTS) {
            ripHdrDisplay(&prip, len);
        }
	msgConstructEmpty(&msg);
        msgPush(&msg, ripHdrStore, (char *)&prip, len,0);
        xPush(udp_s,&msg);
        msgDestroy(&msg);
	return OK;
}

/*------------------------------------------------------------------------
 * ripsendrequest - send a RIP request for entire routing table
 *------------------------------------------------------------------------
 */
int ripsendrequest(self, lls)
XObj		self;
XObj		lls;
{
        Msg		msg;
	struct	rip	prip;
	int		len;

        len = RIPHSIZE + sizeof(struct riprt);

        /* fill in proute */
        bzero((char *)&prip,sizeof(prip));
	prip.rip_rts[0].rr_family = 0;
	prip.rip_rts[0].rr_metric = RIP_INFINITY;

        prip.rip_cmd = RIP_REQUEST;
        prip.rip_vers = RIP_VERSION;
        prip.rip_mbz = 0;
        xIfTrace(ripp, TR_EVENTS) {
            ripHdrDisplay(&prip, len);
        }
        msgConstructEmpty(&msg);
        msgPush(&msg, ripHdrStore, (char *)&prip, len,0);
        xPush(lls,&msg);
        msgDestroy(&msg);
        return OK;
}


/*------------------------------------------------------------------------
 * ripsend - send a RIP update
 *------------------------------------------------------------------------
 */
int ripsend(self, gw, port, udp_s)
XObj		self;
IPhost		gw;		/* remote gateway (FFFFFFFF => all)	*/
unsigned short	port;		/* remote port				*/
XObj		udp_s;
{
        PSTATE 	*pstate = (PSTATE *)self->state;
        RipRouteTable	*rtTbl;
	struct	rq	*prq, rqinfo[RIPMAXINTER];/* here, so reentrant	*/
	struct	route	*prt;
	int		i, pn;
	Msg		msg;

        xTrace0(ripp, TR_EVENTS, "ripsend entered");

	rtTbl = &pstate->rtTbl;

	if (ripifset(self, rqinfo, gw, port, udp_s) != OK)
		return SYSERR;

	for (i=0; i<ROUTETABLESIZE; ++i) {
		for (prt=rtTbl->arr[i]; prt; prt=prt->next) {
			ripadd(self, rqinfo, prt);
		}
	}
	if (rtTbl->defrt) {
		ripadd(self, rqinfo, rtTbl->defrt);
	}

	for (i=0; i<pstate->numNif; ++i) {
		if (rqinfo[i].rq_active) {
			prq = &rqinfo[i];
			for (pn=0; pn<=prq->rq_cur; ++pn) {
                            xIfTrace(ripp, TR_EVENTS) {
                                printf("output packet to interface %d\n",i);
                            }
			    msgConstructEmpty(&msg);
			    msgPush(&msg,ripHdrStore,&prq->rq_prip[pn],prq->rq_len[pn], 0);
			    xPush(prq->rq_udp_s, &msg);
			    msgDestroy(&msg);
			}
		}
	}
        return OK;
}

/*------------------------------------------------------------------------
 * ripifset - set the per-interface data for a RIP update
 *------------------------------------------------------------------------
 */
int ripifset(self, rqinfo, gw, port, udp_s)
XObj		self;
struct	rq	rqinfo[];
IPhost		gw;		/* remote gateway (FFFFFFFF => all)	*/
unsigned short	port;		/* remote port				*/
XObj		udp_s;
{
	PSTATE 	*pstate = (PSTATE *)self->state;
	struct	route	riproute;
	struct	ifc	*ifcptr;
	int		ifn;

                            
        xTrace0(ripp, TR_EVENTS, "ripifset entered");

	if (udp_s) {
		for (ifn=0; ifn<pstate->numNif; ++ifn) {
		        rqinfo[ifn].rq_active = FALSE;
		}
		ifcptr = &pstate->nif[ifn];
               	if (! (ript_get(&pstate->rtTbl,&gw,&riproute) == XK_SUCCESS)) {
                       	return SYSERR;
		}
               	ifn = riproute.ifnum;
		rqinfo[ifn].rq_active = TRUE;
		rqinfo[ifn].rq_udp_s = udp_s;
		rqinfo[ifn].rq_cur = -1;
		rqinfo[ifn].rq_nrts = MAXRIPROUTES;
		return OK;
	}
	/* else, all interfaces */
	for (ifn=0; ifn<pstate->numNif; ++ifn) {
		ifcptr = &pstate->nif[ifn];
		rqinfo[ifn].rq_udp_s = ifcptr->bcast_s;
		rqinfo[ifn].rq_active = TRUE;
		rqinfo[ifn].rq_cur = -1;
		rqinfo[ifn].rq_nrts = MAXRIPROUTES;
	}
	return OK;
}


int ripadd(self, rqinfo, prt)
XObj	self;
struct	rq	rqinfo[];
struct	route	*prt;
{
	PSTATE *pstate = (PSTATE *)self->state;
	IPhost	net;
	int	i, metric, pn, rn;

                            
        xTrace0(ripp, TR_EVENTS, "ripadd entered");

	for (i=0; i<pstate->numNif; ++i) {
		struct	rq	*prq = &rqinfo[i];
		struct	riprt	*rp;

		if (!rqinfo[i].rq_active || !pstate->nif[i].active )
			continue;
		metric = ripmetric(prt, i);
		if (metric == SYSERR)
			continue;
		if (prq->rq_nrts >= MAXRIPROUTES &&
		    ripstart(prq) != OK)
			continue;
		pn = prq->rq_cur;
		rn = prq->rq_nrts++;
		rp = &prq->rq_prip->rip_rts[rn];
		rp->rr_family = AF_INET;
		rp->rr_mbz = 0;
		netnum(&net, &prt->net);
		bzero((char *)&rp->rr_addr, sizeof(rp->rr_addr));
		if ( IP_EQUAL(pstate->nif[i].net, net) ||
		    IP_EQUAL(prt->mask, ip_maskall)) {
			rp->rr_addr = prt->net; 
		} else	/* send the net part only (esp. for subnets)	*/
			rp->rr_addr = net; 
		rp->rr_metric = metric;
		prq->rq_len[pn] += sizeof(struct riprt);
	}
	return OK;
}

/*------------------------------------------------------------------------
 *  netnum  -  compute the network portion of a given IP address
 *------------------------------------------------------------------------
 */
int netnum(net, ipa)
IPhost  *net, *ipa;
{
	IPhost mask;

        xTrace0(ripp, TR_EVENTS, "netnum entered");

	netMaskDefault(&mask,ipa);
	IP_AND(*net,mask,*ipa); 	
        return OK;
}


/*------------------------------------------------------------------------
 * ripstart - initialize an interface's RIP packet data
 *------------------------------------------------------------------------
 */
int ripstart(prq)
struct	rq	*prq;
{
	int		pn;

        xTrace0(ripp, TR_EVENTS, "ripstart entered");

	pn = ++prq->rq_cur;
	if (pn >= MAXNRIP)
		return SYSERR;
	prq->rq_prip = &prq->rq_pep[pn];
	prq->rq_nrts = 0;
	prq->rq_len[pn] = RIPHSIZE;
	prq->rq_prip->rip_cmd = RIP_RESPONSE;
	prq->rq_prip->rip_vers = RIP_VERSION;
	prq->rq_prip->rip_mbz = 0;
	return OK;
}


/*------------------------------------------------------------------------
 *  ripout  -  do the RIP route exchange protocol, output side
 *------------------------------------------------------------------------
 */
void ripout(self)
XObj	self;
{
  	PSTATE 		*pstate;

        xTrace0(ripp, TR_EVENTS, "ripout entered");

  	pstate = (PSTATE *) self->state;
	/* advertise our routes */

	semWait(&pstate->riplock);
	ripsend(self, ip_maskall, UP_RIP, 0);
	semSignal(&pstate->riplock);
}


/*------------------------------------------------------------------------
 * ripmetric - compute the RIP metric for a route we advertise
 *------------------------------------------------------------------------
 */
int ripmetric(prt, ifnum)
struct	route	*prt;
int		ifnum;
{
        xTrace0(ripp, TR_EVENTS, "ripmetric entered");

	if (prt->ifnum == ifnum)
		return RIP_INFINITY;	/* poison reverse */
	/* else, add one to our cost */
	return prt->metric + 1;
}

 
/*
 * ripHdrDisplay --print rip packet 
 */
static void
ripHdrDisplay(hdr,len)
struct rip      *hdr;
int             len;
{
    int num_routes;
    int i;

    printf("RIP Header:\n");
    switch (hdr->rip_cmd) {
    case RIP_REQUEST:
        printf("rip_cmd = RIP_REQUEST\n");
        break;
    case RIP_RESPONSE:
        printf("rip_cmd = RIP_RESPONSE\n");
        break;
    default:
        printf("rip_cmd = Invalid Value %d\n", hdr->rip_cmd);
        break;
    }
    printf("rip_vers = %d rip_mbz = %d\n",hdr->rip_vers,hdr->rip_mbz);

    num_routes =  (len-RIPHSIZE)/sizeof(struct riprt);
 
    printf("number of routes = %d \n",num_routes);
    for (i =0; i< num_routes; i++) {
        printf("[%d] family = %d net =  %s metric = %d \n", i,  hdr->rip_rts[i].rr_family, ipHostStr(&hdr->rip_rts[i].rr_addr), hdr->rip_rts[i].rr_metric);
    }
}
 


/*------------------------------------------------------------------------
 *  ripprint  -  print the routing table
 *------------------------------------------------------------------------
 */ 
void ripprint(self)
XObj	self;
{
        PSTATE          *pstate = (PSTATE *) self->state;
        int             i;
        struct  route   *prt;

        xTrace0(ripp, TR_EVENTS, "ripprint entered");

        semWait(&pstate->riplock);
        for (i=0; i<ROUTETABLESIZE; ++i) {
                for (prt=pstate->rtTbl.arr[i]; prt; prt=prt->next) {
                        ripprintentry(prt);
                }
        }
        semSignal(&pstate->riplock);
}

void ripprintentry(prt)
route *prt;
{   
    printf("net %s mask %s gw %s metric %d \n   sessn %x key %d ttl %d ifnum %d\n", 
    ipHostStr(&prt->net), ipHostStr(&prt->mask), ipHostStr(&prt->gw), 
    prt->metric, prt->netSessn, prt->key, prt->ttl, prt->ifnum);
}   
    
/*------------------------------------------------------------------------
 *  ripage  -  age routes
 *------------------------------------------------------------------------
 */
void ripage(self)
XObj	self;
{
    PSTATE          *pstate = (PSTATE *) self->state;
    RipRouteTable   *tbl = (RipRouteTable *)&(pstate->rtTbl);
    route           *ptr, *prev;
    int             i;
    int		    changed;
   


    xTrace0(ripp, TR_EVENTS, "ripage entered");


    semWait(&pstate->riplock);
    changed = 0;
    for ( i = 0; i < ROUTETABLESIZE; i++) {
        if ( tbl->arr[i] == 0 ) {
            continue;
        }
        prev = NULL;
        for ( ptr = tbl->arr[i]; ptr; ) {
            if ( ptr->ttl != RIPROUTE_TTL_INFINITE ) {
                if (ptr->ttl > RTTABLEDELTA) ptr->ttl -= RTTABLEDELTA;
		else ptr->ttl = 0;
            }
            if ( ptr->ttl == 0 ) {
	
		if (ptr->metric == RIP_INFINITY) {
                    if ( prev ) {
                        prev->next = ptr->next;
                        ript_free(tbl, ptr);
                        ptr = prev->next;
                    } else {
                        tbl->arr[i] = ptr->next;
                        ript_free(tbl, ptr);
                        ptr = tbl->arr[i];
                    }
                    continue;
                } else {
	            ptr->metric = RIP_INFINITY;
	            ptr->ttl = RIPZTIME;
		    changed = 1;
                }
            }
            prev = ptr;
            ptr = ptr->next;
        }
    }

    if (changed && !pstate->outputForced) {
	if ((pstate->outputInterval - pstate->outputTime) > RIPOUTMIN) {
	    pstate->outputTime = 0;
	} else { 
	    pstate->outputTime = RIPOUTMIN;
	}
	pstate->outputForced = 1;
    }

    semSignal(&pstate->riplock);

}

/*------------------------------------------------------------------------
 *  ripclock  -  called every ten seconds to check for stuff to do 
 *------------------------------------------------------------------------
 */
void ripclock(ev, arg)
Event   ev;
VOID    *arg;
{
    XObj            self = (XObj)arg;
    PSTATE          *pstate = (PSTATE *) self->state;
    static 	    rnd = 0;

    xTrace0(ripp, TR_FULL_TRACE, "ripclock entered");

    xIfTrace(ripp, TR_FULL_TRACE) { 
        pstate->printTime -= 1;
        if (pstate->printTime <= 0) {
	    ripprint(self);
	    pstate->printTime = pstate->printInterval;
        }
    }

    pstate->ageTime -= 1;
    if (pstate->ageTime <= 0) {
	ripage(self);
	pstate->ageTime = pstate->ageInterval;
    }

    if (pstate->numNif > 1) {
    	pstate->outputTime -= 1;
        if (pstate->outputTime <= 0) {
            ripout(self);
	    if (++rnd > RIPDELTA) {
		rnd = - RIPDELTA;
	    }
	    pstate->outputTime = pstate->outputInterval + rnd;
            pstate->outputForced = 0;
        }
    }

    evDetach( evSchedule( ripclock, self, ONESEC));
}

