/*
 * tcp_var.h
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)tcp_var.h	7.6 (Berkeley) 12/7/87
 *
 * Modified for x-kernel v3.2
 * Modifications Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1996/07/11 21:39:21 $
 */

/*
 * xtcp_s configuration:  This is a half-assed attempt to make xtcp_s
 * self-configure for a few varieties of 4.2 and 4.3-based unixes.
 * If you don't have a) a 4.3bsd vax or b) a 3.x Sun (x<6), check
 * this carefully (it's probably not right).  Please send me mail
 * if you run into configuration problems.
 *  - Van Jacobson (van@lbl-csam.arpa)
 */

#ifndef BSD
#define BSD 42	/* if we're not 4.3, pretend we're 4.2 */
#endif

#if sun||BSD<43
#define TCP_S_COMPAT_42	/* set if we have to interop w/4.2 systems */
#endif

#ifndef SB_MAX
#ifdef SB_MAXCOUNT
#define	SB_MAX SB_MAXCOUNT	/* Sun has to be a little bit different... */
#else
#define SB_MAX 65535		/* XXX */
#endif /* SB_MAXCOUNT */
#endif /* SB_MAX */

/*
 * Bill Nowicki pointed out that the page size (CLBYTES) has
 * nothing to do with the mbuf cluster size.  So, we followed
 * Sun's lead and made the new define MCLBYTES stand for the mbuf
 * cluster size.  The following define makes up backwards compatible
 * with 4.3 and 4.2.  If CLBYTES is >1024 on your machine, check
 * this against the mbuf cluster definitions in /usr/include/sys/mbuf.h.
 */
#ifndef MCLBYTES
#define	MCLBYTES CLBYTES	/* XXX */
#endif

/*
 * The routine in_localaddr is broken in Sun's 3.4.  We redefine ours
 * (in tcp_s_input.c) so we use can it but won't have a name conflict.
 */
#ifdef sun
#define in_localaddr tcp_s_in_localaddr
#endif

/* --------------- end of xtcp_s config ---------------- */

/*
 * Kernel variables for tcp_s.
 */

struct reass {
  struct reass *next, *prev;
  struct tcp_shdr th;
  Msg m;
};

/*
 * Tcp_S control block, one per tcp_s; fields:
 */
struct tcp_scb {
	struct	reass *seg_next;	/* sequencing queue */
	struct	reass *seg_prev;
	short	t_state;		/* state of this connection */
	short	t_timer[TCP_ST_NTIMERS];	/* tcp_s timers */
	short	t_rxtshift;		/* log(2) of rexmt exp. backoff */
	short	t_rxtcur;		/* current retransmit value */
	short	t_dupacks;		/* consecutive dup acks recd */
	u_short	t_maxseg;		/* maximum segment size */
	char	t_force;		/* 1 if forcing out a byte */
	u_char	t_flags;
#define	TF_ACKNOW	0x01		/* ack peer immediately */
#define	TF_DELACK	0x02		/* ack, but try to delay it */
#define	TF_NODELAY	0x04		/* don't delay packets to coalesce */
#define	TF_NOOPT	0x08		/* don't use tcp_s options */
#define	TF_SENTFIN	0x10		/* have sent FIN */
	struct	tcp_siphdr *t_template;	/* skeletal packet for transmit */
	struct	inpcb *t_inpcb;		/* back pointer to internet pcb */
/*
 * The following fields are used as in the protocol specification.
 * See RFC783, Dec. 1981, page 21.
 */
/* send sequence variables */
	tcp_s_seq	snd_una;		/* send unacknowledged */
	tcp_s_seq	snd_nxt;		/* send next */
	tcp_s_seq	snd_up;			/* send urgent pointer */
	tcp_s_seq	snd_wl1;		/* window update seg seq number */
	tcp_s_seq	snd_wl2;		/* window update seg ack number */
	tcp_s_seq	iss;			/* initial send sequence number */
	u_short	snd_wnd;		/* send window */
/* receive sequence variables */
	u_short	rcv_wnd;		/* receive window */
	tcp_s_seq	rcv_nxt;		/* receive next */
	tcp_s_seq	rcv_up;			/* receive urgent pointer */
	tcp_s_seq	irs;			/* initial receive sequence number */
/*
 * Additional variables for this implementation.
 */
/* receive variables */
	tcp_s_seq	rcv_adv;		/* advertised window */
/* retransmit variables */
	tcp_s_seq	snd_max;		/* highest sequence number sent
					 * used to recognize retransmits
					 */
/* congestion control (for slow start, source quench, retransmit after loss) */
	u_short	snd_cwnd;		/* congestion-controlled window */
	u_short snd_ssthresh;		/* snd_cwnd size threshhold for
					 * for slow start exponential to
					 * linear switch */
/*
 * transmit timing stuff.
 * srtt and rttvar are stored as fixed point; for convenience in smoothing,
 * srtt has 3 bits to the right of the binary point, rttvar has 2.
 * "Variance" is actually smoothed difference.
 */
	short	t_idle;			/* inactivity time */
	short	t_rtt;			/* round trip time */
	tcp_s_seq	t_rtseq;		/* sequence number being timed */
	short	t_srtt;			/* smoothed round-trip time */
	short	t_rttvar;		/* variance in round-trip time */
	u_short max_rcvd;		/* most peer has sent into window */
	u_short	max_sndwnd;		/* largest window peer has offered */
/* out-of-band data */
	char	t_oobflags;		/* have some */
	char	t_iobc;			/* input character */
#define	TCP_SOOB_HAVEDATA	0x01
#define	TCP_SOOB_HADDATA	0x02
};

#define	intotcp_scb(ip)	((struct tcp_scb *)(ip)->inp_ppcb)
#define	sototcp_scb(so)	(intotcp_scb(sotoinpcb(so)))

/*
 * TCP_S statistics.
 * Many of these should be kept per connection,
 * but that's inconvenient at the moment.
 */
struct	tcp_sstat {
#if BSD<=43
       int     tcp_ss_badsum;
       int     tcp_ss_badoff;
       int     tcp_ss_hdrops;
       int     tcp_ss_badsegs;
       int     tcp_ss_unack;
       /* 4.3+ BSD stats start here */
#endif
	u_long	tcp_ss_connattempt;	/* connections initiated */
	u_long	tcp_ss_accepts;		/* connections accepted */
	u_long	tcp_ss_connects;		/* connections established */
	u_long	tcp_ss_drops;		/* connections dropped */
	u_long	tcp_ss_conndrops;		/* embryonic connections dropped */
	u_long	tcp_ss_closed;		/* conn. closed (includes drops) */
	u_long	tcp_ss_segstimed;		/* segs where we tried to get rtt */
	u_long	tcp_ss_rttupdated;	/* times we succeeded */
	u_long	tcp_ss_delack;		/* delayed acks sent */
	u_long	tcp_ss_timeoutdrop;	/* conn. dropped in rxmt timeout */
	u_long	tcp_ss_rexmttimeo;	/* retransmit timeouts */
	u_long	tcp_ss_persisttimeo;	/* persist timeouts */
	u_long	tcp_ss_keeptimeo;		/* keepalive timeouts */
	u_long	tcp_ss_keepprobe;		/* keepalive probes sent */
	u_long	tcp_ss_keepdrops;		/* connections dropped in keepalive */

	u_long	tcp_ss_sndtotal;		/* total packets sent */
	u_long	tcp_ss_sndpack;		/* data packets sent */
	u_long	tcp_ss_sndbyte;		/* data bytes sent */
	u_long	tcp_ss_sndrexmitpack;	/* data packets retransmitted */
	u_long	tcp_ss_sndrexmitbyte;	/* data bytes retransmitted */
	u_long	tcp_ss_sndacks;		/* ack-only packets sent */
	u_long	tcp_ss_sndprobe;		/* window probes sent */
	u_long	tcp_ss_sndurg;		/* packets sent with URG only */
	u_long	tcp_ss_sndwinup;		/* window update-only packets sent */
	u_long	tcp_ss_sndctrl;		/* control (SYN|FIN|RST) packets sent */

	u_long	tcp_ss_rcvtotal;		/* total packets received */
	u_long	tcp_ss_rcvpack;		/* packets received in sequence */
	u_long	tcp_ss_rcvbyte;		/* bytes received in sequence */
	u_long	tcp_ss_rcvbadsum;		/* packets received with ccksum errs */
	u_long	tcp_ss_rcvbadoff;		/* packets received with bad offset */
	u_long	tcp_ss_rcvshort;		/* packets received too short */
	u_long	tcp_ss_rcvduppack;	/* duplicate-only packets received */
	u_long	tcp_ss_rcvdupbyte;	/* duplicate-only bytes received */
	u_long	tcp_ss_rcvpartduppack;	/* packets with some duplicate data */
	u_long	tcp_ss_rcvpartdupbyte;	/* dup. bytes in part-dup. packets */
	u_long	tcp_ss_rcvoopack;		/* out-of-order packets received */
	u_long	tcp_ss_rcvoobyte;		/* out-of-order bytes received */
	u_long	tcp_ss_rcvpackafterwin;	/* packets with data after window */
	u_long	tcp_ss_rcvbyteafterwin;	/* bytes rcvd after window */
	u_long	tcp_ss_rcvafterclose;	/* packets rcvd after "close" */
	u_long	tcp_ss_rcvwinprobe;	/* rcvd window probe packets */
	u_long	tcp_ss_rcvdupack;		/* rcvd duplicate acks */
	u_long	tcp_ss_rcvacktoomuch;	/* rcvd acks for unsent data */
	u_long	tcp_ss_rcvackpack;	/* rcvd ack packets */
	u_long	tcp_ss_rcvackbyte;	/* bytes acked by rcvd acks */
	u_long	tcp_ss_rcvwinupd;		/* rcvd window update packets */
};


#ifdef __STDC__

struct	tcp_siphdr *tcp_s_template( struct tcp_scb * );
struct	tcp_scb *tcp_s_destroy( struct tcp_scb * );
struct  tcp_scb *tcp_s_drop( struct tcp_scb *, int );
struct	tcp_scb *tcp_s_timers( struct tcp_scb *, int );
struct  tcp_scb *tcp_s_disconnect( struct tcp_scb * );
struct  tcp_scb *tcp_s_usrclosed( struct tcp_scb * );

#else

struct	tcp_siphdr *tcp_s_template();
struct	tcp_scb *tcp_s_destroy();
struct  tcp_scb *tcp_s_drop();
struct	tcp_scb *tcp_s_timers();
struct  tcp_scb *tcp_s_disconnect();
struct  tcp_scb *tcp_s_usrclosed();

#endif /* __STDC__ */
