/*
 * $RCSfile: config.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: config.c,v $
 * Revision 1.1  1994/10/13 17:43:30  dorgival
 * Initial revision
 *
 */

/*
 * Target compiler "configuration": handles information about basic types
 * in the target compiler based on pragmas.
 */

#include "usc.h"

#define BUFFER_SIZE 256

extern int legal_byte_order (ByteOrder string);

int configured = 0;

MachineInfo machine_info = {
  DEFAULT_NUM_REGS,
  DEFAULT_REG_SIZE,
  {
    DEFAULT_SIZE,
    -1,                       /* DEFAULT_LONG_MEMSIZE,    */
    DEFAULT_ALIGN,
    DEFAULT_BORDER,           /* DEFAULT_LONG_BYTEORDER,  */
    "long"
  },
  {
    DEFAULT_SIZE,
    -1,                       /* DEFAULT_INT_MEMSIZE,     */
    DEFAULT_ALIGN,
    DEFAULT_BORDER,            /* DEFAULT_INT_BYTEORDER,   */
    "int"
  },
  {
    DEFAULT_SIZE,
    -1,                       /* DEFAULT_SHORT_MEMSIZE,   */
    DEFAULT_ALIGN,
    DEFAULT_BORDER,           /* DEFAULT_SHORT_BYTEORDER, */
    "short"
  },
  {
    DEFAULT_SIZE,
    -1,                       /* DEFAULT_CHAR_MEMSIZE,    */
    DEFAULT_ALIGN,
    DEFAULT_BORDER,           /* DEFAULT_CHAR_BYTEORDER,  */
    "char"
  }
};

void print_config (int level)
{
  if (level <= Debug_Level) {
    printf ("long size: %d\n", machine_info.long_info.size);
    printf ("long alignment: %d\n", machine_info.long_info.alignment);
    printf ("long byteorder: %s\n",
	     byteorder_to_string(machine_info.long_info.byteorder));
    printf ("int size: %d\n", machine_info.int_info.size);
    printf ("int alignment: %d\n", machine_info.int_info.alignment);
    printf ("int byteorder: %s\n",
	     byteorder_to_string(machine_info.int_info.byteorder));
    printf ("short size: %d\n", machine_info.short_info.size);
    printf ("short alignment: %d\n", machine_info.short_info.alignment);
    printf ("short byteorder: %s\n",
	     byteorder_to_string(machine_info.short_info.byteorder));
    printf ("char size: %d\n", machine_info.char_info.size);
    printf ("char alignment: %d\n", machine_info.char_info.alignment);
    printf ("char byteorder: %s\n",
	     byteorder_to_string(machine_info.char_info.byteorder));
  }
}

void check_config (void)
{
  int fail = 0;

  if ( do_preprocess )
    return;
  if (machine_info.long_info.memsize <= 0) {
    fprintf (stderr, "long native type not initialized\n");
    fail = 1;
  }
  if (machine_info.int_info.memsize <= 0) {
    fprintf (stderr, "int native type not initialized\n");
    fail = 1;
  }
  if (machine_info.short_info.memsize <= 0) {
    fprintf (stderr, "short native type not initialized\n");
    fail = 1;
  }
  if (machine_info.char_info.memsize <= 0) {
    fprintf (stderr, "char native type not initialized\n");
    fail = 1;
  }

  if (machine_info.long_info.size <= 0) {
    fprintf (stderr, "Invalid tsize for long native type\n");
    fail = 1;
  }
  if (machine_info.int_info.size <= 0) {
    fprintf (stderr, "Invalid tsize for int native type\n");
    fail = 1;
  }
  if (machine_info.short_info.size <= 0) {
    fprintf (stderr, "Invalid tsize for short native type\n");
    fail = 1;
  }
  if (machine_info.char_info.size <= 0) {
    fprintf (stderr, "Invalid tsize for char native type\n");
    fail = 1;
  }

  if (machine_info.long_info.alignment <= 0) {
    fprintf (stderr, "Invalid alignment for long native type\n");
    fail = 1;
  }
  if (machine_info.int_info.alignment <= 0) {
    fprintf (stderr, "Invalid alignment for int native type\n");
    fail = 1;
  }
  if (machine_info.short_info.alignment <= 0) {
    fprintf (stderr, "Invalid alignment for short native type\n");
    fail = 1;
  }
  if (machine_info.char_info.alignment <= 0) {
    fprintf (stderr, "Invalid alignment for char native type\n");
    fail = 1;
  }

  if (!machine_info.long_info.byteorder||
      !legal_byte_order (machine_info.long_info.byteorder) ||
      byteorder_len (machine_info.long_info.byteorder) != 
      machine_info.long_info.size) {
    fprintf (stderr, "long byte order not properly initialized\n");
    fail = 1;
  }

  if (!machine_info.int_info.byteorder ||
      !legal_byte_order (machine_info.int_info.byteorder) ||
      byteorder_len (machine_info.int_info.byteorder) != 
      machine_info.int_info.size) {
    fprintf (stderr, "int byte order not properly initialized\n");
    fail = 1;
  }

  if (!machine_info.short_info.byteorder ||
      !legal_byte_order (machine_info.short_info.byteorder) ||
      byteorder_len (machine_info.short_info.byteorder) != 
      machine_info.short_info.size){
    fprintf (stderr, "short byte order not properly initialized\n");
    fail = 1;
  }

  if (!machine_info.char_info.byteorder ||
      !legal_byte_order (machine_info.char_info.byteorder) ||
      byteorder_len (machine_info.char_info.byteorder) != 
      machine_info.char_info.size) {
    fprintf (stderr, "char byte order not properly initialized\n");
    fail = 1;
  }
  if (fail)
    exit (1);
  else
    configured = 1;
}
